from datetime import date
from morepath import Response
from onegov.core.security import Public
from onegov.core.templates import render_macro
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import MunicipalityIdSelectionForm
from onegov.wtfs.layouts import DefaultLayout
from onegov.wtfs.models import PickupDate
from onegov.wtfs.models import Principal
from onegov.wtfs.security import ViewModel


@WtfsApp.html(
    model=Principal,
    template='home.pt',
    permission=Public
)
def view_home(self, request):
    """ The home page. """

    layout = DefaultLayout(self, request)

    if not request.is_logged_in:
        return request.redirect(layout.login_url)

    if not layout.notifications.query().first():
        return request.redirect(layout.top_navigation[0].attrs['href'])

    return {'layout': layout}


@WtfsApp.form(
    model=Principal,
    permission=ViewModel,
    name='dispatch-dates',
    form=MunicipalityIdSelectionForm
)
def view_dispatch_dates(self, request, form):
    if form.submitted(request):
        layout = DefaultLayout(self, request)
        dates = [
            r.date for r in form.municipality.pickup_dates.filter(
                PickupDate.date > date.today()
            )
        ]
        return render_macro(
            layout.macros['dispatch_dates'],
            request,
            {'dates': dates, 'layout': layout}
        )

    return Response()


@WtfsApp.form(
    model=Principal,
    permission=ViewModel,
    name='price-per-quantity',
    form=MunicipalityIdSelectionForm
)
def view_price_per_quantity(self, request, form):
    if form.submitted(request):
        layout = DefaultLayout(self, request)
        price_per_quantity = form.municipality.price_per_quantity
        return render_macro(
            layout.macros['price_per_quantity'],
            request,
            {'price_per_quantity': price_per_quantity, 'layout': layout}
        )

    return Response()



from cached_property import cached_property
from onegov.wtfs.layouts.default import DefaultLayout


class MailLayout(DefaultLayout):
    """ A special layout for creating HTML E-Mails. """

    @cached_property
    def base(self):
        return self.template_loader['mail_layout.pt']

    @cached_property
    def primary_color(self):
        return self.app.theme_options.get('primary-color', '#fff')



from onegov.wtfs.fields.csv import CsvUploadField
from onegov.wtfs.fields.hint import HintField
from onegov.wtfs.fields.municipality import MunicipalityDataUploadField


__all__ = (
    'CsvUploadField',
    'HintField',
    'MunicipalityDataUploadField',
)



from cgi import FieldStorage
from datetime import date
from freezegun import freeze_time
from io import BytesIO
from mock import MagicMock
from onegov.user import User
from onegov.user import UserCollection
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.forms import AddScanJobForm
from onegov.wtfs.forms import CreateInvoicesForm
from onegov.wtfs.forms import DailyListSelectionForm
from onegov.wtfs.forms import DeleteMunicipalityDatesForm
from onegov.wtfs.forms import EditScanJobForm
from onegov.wtfs.forms import ImportMunicipalityDataForm
from onegov.wtfs.forms import MunicipalityForm
from onegov.wtfs.forms import MunicipalityIdSelectionForm
from onegov.wtfs.forms import NotificationForm
from onegov.wtfs.forms import PaymentTypesForm
from onegov.wtfs.forms import ReportSelectionForm
from onegov.wtfs.forms import ScanJobsForm
from onegov.wtfs.forms import UnrestrictedScanJobForm
from onegov.wtfs.forms import UnrestrictedScanJobsForm
from onegov.wtfs.forms import UnrestrictedUserForm
from onegov.wtfs.forms import UserForm
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Notification
from onegov.wtfs.models import PickupDate
from onegov.wtfs.models import ScanJob


class App(object):
    def __init__(self, session, principal):
        self._session = session
        self.principal = principal

    def session(self):
        return self._session


class Identity(object):
    def __init__(self, groupid):
        self.groupid = groupid


class Request(object):
    def __init__(self, session, principal=None, groupid=None, roles=[]):
        self.app = App(session, principal)
        self.session = session
        self.identity = Identity(groupid)
        self.roles = roles

    def include(self, resource):
        pass

    def has_role(self, role):
        return role in self.roles

    def translate(self, text):
        return text.interpolate()


class PostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


def test_payments_form(session):
    model = PaymentTypeCollection(session)
    request = Request(session)

    form = PaymentTypesForm.get_form_class(model, request)()
    assert [field for field in form] == []

    model.add(name='normal', _price_per_quantity=700)
    model.add(name='spezial', _price_per_quantity=850)
    form = PaymentTypesForm.get_form_class(model, request)()
    assert [(field.name, field.type) for field in form] == [
        ('normal', 'FloatField'), ('spezial', 'FloatField')
    ]

    # Test apply / update
    form.apply_model(model)
    assert form.normal.data == 7.0
    assert form.spezial.data == 8.5

    form.normal.data = 8.0
    form.spezial.data = 9

    form.update_model(model)
    assert {r.name: r.price_per_quantity for r in model.query()} == {
        'normal': 8.0, 'spezial': 9.0
    }

    # Test validation
    form = PaymentTypesForm.get_form_class(model, request)(
        PostData({
            'normal': 10.0,
            'spezial': 12.6,
        })
    )
    form.request = Request(session)
    assert form.validate()


def test_municipality_form(session):
    payment_types = PaymentTypeCollection(session)
    payment_types.add(name='normal', _price_per_quantity=700)
    payment_types.add(name='spezial', _price_per_quantity=850)

    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(
        name="Boppelsen",
        bfs_number=82,
        payment_type='normal'
    )

    # Test choices
    form = MunicipalityForm()
    form.request = Request(session)
    form.on_request()
    assert form.payment_type.choices == [
        ('normal', 'Normal'), ('spezial', 'Spezial')
    ]

    # Test apply / update
    form.apply_model(municipality)
    assert form.name.data == "Boppelsen"
    assert form.bfs_number.data == 82
    assert form.address_supplement.data is None
    assert form.gpn_number.data is None
    assert form.payment_type.data == 'normal'

    form.name.data = "Adlikon"
    form.bfs_number.data = 21
    form.address_supplement.data = "Zusatz"
    form.gpn_number.data = 1122
    form.payment_type.data = 'spezial'

    form.update_model(municipality)
    assert municipality.name == "Adlikon"
    assert municipality.bfs_number == 21
    assert municipality.address_supplement == "Zusatz"
    assert municipality.gpn_number == 1122
    assert municipality.payment_type == 'spezial'

    # Test validation
    form = MunicipalityForm()
    form.request = Request(session)
    form.on_request()
    assert not form.validate()

    form = MunicipalityForm(
        PostData({
            'bfs_number': '82',
        })
    )
    form.request = Request(session)
    form.on_request()
    assert not form.validate()

    form = MunicipalityForm(
        PostData({
            'name': "Boppelsen",
            'bfs_number': '82',
            'payment_type': 'normal'
        })
    )
    form.request = Request(session)
    form.on_request()
    assert form.validate()


def test_import_municipality_data_form(session):
    municipalities = MunicipalityCollection(session)
    municipalities.add(name="Boppelsen", bfs_number=82)
    municipalities.add(name="Adlikon", bfs_number=21)

    # Test apply
    form = ImportMunicipalityDataForm()
    form.request = Request(session)

    form.file.data = {
        21: {'dates': [date(2019, 1, 1), date(2019, 1, 7)]},
        241: {'dates': [date(2019, 1, 3), date(2019, 1, 9)]},
        82: {'dates': [date(2019, 1, 4), date(2019, 1, 10)]}
    }
    form.update_model(municipalities)
    assert [
        (m.bfs_number, [d.date for d in m.pickup_dates])
        for m in municipalities.query()
    ] == [
        (21, [date(2019, 1, 1), date(2019, 1, 7)]),
        (82, [date(2019, 1, 4), date(2019, 1, 10)])
    ]

    # Test validation
    form = ImportMunicipalityDataForm()
    form.request = Request(session)
    assert not form.validate()

    field_storage = FieldStorage()
    field_storage.file = BytesIO(
        "Gemeinde-Nr,Vordefinierte Termine\n21,12.2.2015".encode('utf-8')
    )
    field_storage.type = 'text/csv'
    field_storage.filename = 'test.csv'
    form.file.process(PostData({'file': field_storage}))

    assert form.validate()


def test_delete_municipality_dates_form(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name="Boppelsen", bfs_number=82)

    # Test apply / update
    form = DeleteMunicipalityDatesForm()
    form.request = Request(session)
    form.apply_model(municipality)
    assert form.start.data is None
    assert form.end.data is None

    form.start.data = date(2019, 1, 1)
    form.end.data = date(2019, 12, 31)
    form.update_model(municipality)

    municipality.pickup_dates.append(PickupDate(date=date(2018, 12, 31)))
    municipality.pickup_dates.append(PickupDate(date=date(2019, 1, 1)))
    municipality.pickup_dates.append(PickupDate(date=date(2019, 6, 6)))
    municipality.pickup_dates.append(PickupDate(date=date(2019, 12, 31)))
    municipality.pickup_dates.append(PickupDate(date=date(2020, 1, 1)))

    form.apply_model(municipality)
    assert form.start.data == date(2018, 12, 31)
    assert form.end.data == date(2020, 1, 1)

    form.start.data = date(2019, 1, 1)
    form.end.data = date(2019, 12, 31)
    form.update_model(municipality)
    assert [d.date for d in municipality.pickup_dates] == [
        date(2018, 12, 31), date(2020, 1, 1)
    ]

    # Test validation
    form = DeleteMunicipalityDatesForm()
    assert not form.validate()

    form = DeleteMunicipalityDatesForm(
        PostData({'start': "2019-01-01", 'end': "2019-12-31"})
    )
    assert form.validate()


def test_municipality_selection_form(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name="Aesch", bfs_number=82)

    # Test choices
    form = MunicipalityIdSelectionForm()
    form.request = Request(session)
    form.on_request()
    assert [c[1] for c in form.municipality_id.choices] == ["Aesch"]

    form.municipality_id.data = municipality.id
    assert form.municipality == municipality


def test_user_form(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name="Aesch", bfs_number=82)

    # Test apply / update
    form = UserForm()
    form.request = Request(session, groupid=municipality.id.hex)

    user = User()
    form.realname.data = "Petra Muster"
    form.username.data = "petra.muster@winterthur.ch"
    form.contact.data = False
    form.update_model(user)
    assert user.realname == "Petra Muster"
    assert user.username == "petra.muster@winterthur.ch"
    assert user.group_id == municipality.id.hex
    assert user.data['contact'] is False
    assert user.password_hash
    assert user.modified

    users = UserCollection(session)
    user = users.add(
        realname="Hans Muster",
        username="hans.muster@winterthur.ch",
        role='invalid',
        password='abcd',
    )
    user.logout_all_sessions = MagicMock()
    password_hash = user.password_hash

    form.apply_model(user)
    assert form.realname.data == "Hans Muster"
    assert form.username.data == "hans.muster@winterthur.ch"
    assert form.contact.data is False

    form.realname.data = "Hans-Peter Muster"
    form.username.data = "hans-peter.muster@winterthur.ch"
    form.contact.data = True
    form.update_model(user)
    assert user.realname == "Hans-Peter Muster"
    assert user.username == "hans-peter.muster@winterthur.ch"
    assert user.group_id == municipality.id.hex
    assert user.data['contact'] is True
    assert user.password_hash == password_hash
    assert user.logout_all_sessions.called is True

    # Test validation
    form = UserForm()
    form.request = Request(session, groupid=municipality.id.hex)
    assert not form.validate()

    form = UserForm(PostData({
        'realname': "Hans-Peter Muster",
        'username': "hans-peter.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality.id.hex)
    assert not form.validate()
    assert form.errors == {'username': ['This value already exists.']}

    form = UserForm(PostData({
        'realname': "Hans Muster",
        'username': "hans.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality.id.hex)
    assert form.validate()


def test_unrestricted_user_form(session):
    municipalities = MunicipalityCollection(session)
    municipality_1 = municipalities.add(name="Aesch", bfs_number=82)
    municipality_2 = municipalities.add(name="Adlikon", bfs_number=21)

    # Test choices
    form = UnrestrictedUserForm()
    form.request = Request(session)
    form.on_request()
    assert [c[1] for c in form.municipality_id.choices] == [
        "- none -", "Adlikon (21)", "Aesch (82)"
    ]
    assert form.role.choices == [
        ('editor', "Editor"),
        ('member', "Member")
    ]

    # Test apply / update
    user = User()
    form.role.data = "member"
    form.municipality_id.data = None
    form.realname.data = "Petra Muster"
    form.username.data = "petra.muster@winterthur.ch"
    form.contact.data = False
    form.update_model(user)
    assert user.role == 'member'
    assert user.realname == "Petra Muster"
    assert user.username == "petra.muster@winterthur.ch"
    assert user.group_id is None
    assert user.data['contact'] is False
    assert user.password_hash
    assert user.modified

    users = UserCollection(session)
    user = users.add(
        realname="Hans Muster",
        username="hans.muster@winterthur.ch",
        role='member',
        password='abcd',
    )
    user.group_id = municipality_1.id
    user.logout_all_sessions = MagicMock()
    password_hash = user.password_hash

    form.apply_model(user)
    assert form.role.data == 'member'
    assert form.municipality_id.data == municipality_1.id.hex
    assert form.realname.data == "Hans Muster"
    assert form.username.data == "hans.muster@winterthur.ch"
    assert form.contact.data is False

    form.role.data = 'admin'
    form.municipality_id.data = municipality_2.id.hex
    form.realname.data = "Hans-Peter Muster"
    form.username.data = "hans-peter.muster@winterthur.ch"
    form.contact.data = True
    form.update_model(user)
    assert user.realname == "Hans-Peter Muster"
    assert user.username == "hans-peter.muster@winterthur.ch"
    assert user.group_id == municipality_2.id.hex
    assert user.data['contact'] is True
    assert user.password_hash == password_hash
    assert user.logout_all_sessions.called is True

    # Test validation
    form = UnrestrictedUserForm()
    form.request = Request(session, groupid=municipality_1.id.hex)
    form.on_request()
    assert not form.validate()

    form = UnrestrictedUserForm(PostData({
        'role': 'member',
        'realname': "Hans-Peter Muster",
        'username': "hans-peter.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality_1.id.hex)
    form.on_request()
    assert not form.validate()
    assert form.errors == {'username': ['This value already exists.']}

    form = UnrestrictedUserForm(PostData({
        'role': 'admin',
        'realname': "Hans Muster",
        'username': "hans.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality_1.id.hex)
    form.on_request()
    assert not form.validate()
    assert form.errors == {'role': ['Not a valid choice']}

    form = UnrestrictedUserForm(PostData({
        'role': 'member',
        'realname': "Hans Muster",
        'username': "hans.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality_1.id.hex)
    form.on_request()
    assert form.validate()

    form = UnrestrictedUserForm(PostData({
        'role': 'editor',
        'municipality_id': municipality_2.id.hex,
        'realname': "Hans Muster",
        'username': "hans.muster@winterthur.ch"
    }))
    form.request = Request(session, groupid=municipality_1.id.hex)
    form.on_request()
    assert form.validate()


def test_add_scan_job_form(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name="Boppelsen", bfs_number=82)
    municipality.pickup_dates.append(PickupDate(date=date(2019, 1, 7)))
    municipality.pickup_dates.append(PickupDate(date=date(2019, 1, 8)))

    # Test on request
    form = AddScanJobForm()
    form.request = Request(session, groupid=municipality.id.hex)
    assert form.municipality_id == municipality.id.hex
    with freeze_time("2019-01-05"):
        form.on_request()
        assert form.type.choices == [('normal', 'Regular shipment')]
        assert form.dispatch_date_normal.choices == [
            (date(2019, 1, 7), '07.01.2019'),
            (date(2019, 1, 8), '08.01.2019')
        ]
    form.request = Request(session, groupid=municipality.id.hex,
                           roles=['editor'])
    with freeze_time("2019-01-05"):
        form.on_request()
        assert form.type.choices == [
            ('normal', 'Regular shipment'),
            ('express', 'Express shipment')
        ]
        assert form.dispatch_date_normal.choices == [
            (date(2019, 1, 7), '07.01.2019'),
            (date(2019, 1, 8), '08.01.2019')
        ]

    # Test update
    form.type.data = 'normal'
    form.dispatch_date_normal.data = date(2019, 1, 8)
    form.dispatch_date_express.data = date(2019, 1, 6)
    form.dispatch_boxes.data = 1
    form.dispatch_tax_forms_current_year.data = 2
    form.dispatch_tax_forms_last_year.data = 3
    form.dispatch_tax_forms_older.data = 4
    form.dispatch_single_documents.data = 5
    form.dispatch_note.data = 'Note on dispatch'
    form.dispatch_cantonal_tax_office.data = 6
    form.dispatch_cantonal_scan_center.data = 7

    model = ScanJob()
    form.update_model(model)
    assert model.municipality_id == municipality.id.hex
    assert model.type == 'normal'
    assert model.dispatch_date == date(2019, 1, 8)
    assert model.dispatch_boxes == 1
    assert model.dispatch_tax_forms_current_year == 2
    assert model.dispatch_tax_forms_last_year == 3
    assert model.dispatch_tax_forms_older == 4
    assert model.dispatch_single_documents == 5
    assert model.dispatch_note == 'Note on dispatch'
    assert model.dispatch_cantonal_tax_office == 6
    assert model.dispatch_cantonal_scan_center == 7

    form.type.data = 'express'
    form.update_model(model)
    assert model.dispatch_date == date(2019, 1, 6)

    # Test validation
    with freeze_time("2019-01-05"):
        form = AddScanJobForm()
        form.request = Request(session, groupid=municipality.id.hex)
        form.on_request()
        assert not form.validate()

        form = AddScanJobForm(PostData({
            'municipality_id': municipality.id.hex,
            'type': 'normal',
            'dispatch_date_normal': '2019-01-08'
        }))
        form.request = Request(session, groupid=municipality.id.hex)
        form.on_request()
        assert form.validate()


def test_edit_scan_job_form(session):
    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name="Boppelsen", bfs_number=82)
    municipality.pickup_dates.append(PickupDate(date=date(2019, 1, 7)))
    municipality.pickup_dates.append(PickupDate(date=date(2019, 1, 8)))

    # Test apply / update
    model = ScanJob()
    model.municipality_id = municipality.id
    model.type = 'normal'
    model.dispatch_date = date(2019, 1, 8)
    model.dispatch_boxes = 1
    model.dispatch_tax_forms_current_year = 2
    model.dispatch_tax_forms_last_year = 3
    model.dispatch_tax_forms_older = 4
    model.dispatch_single_documents = 5
    model.dispatch_note = 'Note on dispatch'
    model.dispatch_cantonal_tax_office = 6
    model.dispatch_cantonal_scan_center = 7

    form = EditScanJobForm()
    form.request = Request(session, groupid=municipality.id.hex)
    form.apply_model(model)
    assert form.dispatch_boxes.data == 1
    assert form.dispatch_tax_forms_current_year.data == 2
    assert form.dispatch_tax_forms_last_year.data == 3
    assert form.dispatch_tax_forms_older.data == 4
    assert form.dispatch_single_documents.data == 5
    assert form.dispatch_note.data == 'Note on dispatch'
    assert form.dispatch_cantonal_tax_office.data == 6
    assert form.dispatch_cantonal_scan_center.data == 7

    form.dispatch_boxes.data = 10
    form.dispatch_tax_forms_current_year.data = 20
    form.dispatch_tax_forms_last_year.data = 30
    form.dispatch_tax_forms_older.data = 40
    form.dispatch_single_documents.data = 50
    form.dispatch_note.data = 'A note on the dispatch'
    form.dispatch_cantonal_tax_office.data = 60
    form.dispatch_cantonal_scan_center.data = 70

    form.update_model(model)
    assert model.municipality_id == municipality.id
    assert model.type == 'normal'
    assert model.dispatch_date == date(2019, 1, 8)
    assert model.dispatch_boxes == 10
    assert model.dispatch_tax_forms_current_year == 20
    assert model.dispatch_tax_forms_last_year == 30
    assert model.dispatch_tax_forms_older == 40
    assert model.dispatch_single_documents == 50
    assert model.dispatch_note == 'A note on the dispatch'
    assert model.dispatch_cantonal_tax_office == 60
    assert model.dispatch_cantonal_scan_center == 70

    # Test validation
    form = EditScanJobForm()
    form.request = Request(session, groupid=municipality.id.hex)
    assert form.validate()


def test_unrestricted_scan_job_form(session):
    municipalities = MunicipalityCollection(session)
    municipality_1 = municipalities.add(name="Boppelsen", bfs_number=82)
    municipality_1.pickup_dates.append(PickupDate(date=date(2019, 1, 7)))
    municipality_1.pickup_dates.append(PickupDate(date=date(2019, 1, 8)))

    municipality_2 = municipalities.add(name="Adlikon", bfs_number=21)
    municipality_2.pickup_dates.append(PickupDate(date=date(2019, 1, 17)))
    municipality_2.pickup_dates.append(PickupDate(date=date(2019, 1, 18)))

    # Test on request
    form = UnrestrictedScanJobForm()
    form.request = Request(session, groupid=municipality_1.id.hex)
    with freeze_time("2019-01-05"):
        form.on_request()
        assert form.type.choices == [
            ('normal', 'Regular shipment'),
            ('express', 'Express shipment')
        ]
        assert form.municipality_id.choices == [
            (municipality_2.id.hex, 'Adlikon (21)'),
            (municipality_1.id.hex, 'Boppelsen (82)')
        ]

    # Test apply / update
    model = ScanJob()
    model.municipality_id = municipality_1.id
    model.type = 'normal'
    model.dispatch_date = date(2019, 1, 8)
    model.dispatch_boxes = 1
    model.dispatch_tax_forms_current_year = 2
    model.dispatch_tax_forms_last_year = 3
    model.dispatch_tax_forms_older = 4
    model.dispatch_single_documents = 5
    model.dispatch_note = 'Note on dispatch'
    model.dispatch_cantonal_tax_office = 6
    model.dispatch_cantonal_scan_center = 7
    model.return_date = date(2019, 1, 10)
    model.return_boxes = 8
    model.return_tax_forms_current_year = 9
    model.return_tax_forms_last_year = 10
    model.return_tax_forms_older = 11
    model.return_single_documents = 12
    model.return_unscanned_tax_forms_current_year = 13
    model.return_unscanned_tax_forms_last_year = 14
    model.return_unscanned_tax_forms_older = 15
    model.return_unscanned_single_documents = 16
    model.return_note = 'Note on return'

    form.apply_model(model)
    assert form.type.data == 'normal'
    assert form.municipality_id.data == municipality_1.id.hex
    assert form.dispatch_date.data == date(2019, 1, 8)
    assert form.dispatch_boxes.data == 1
    assert form.dispatch_tax_forms_current_year.data == 2
    assert form.dispatch_tax_forms_last_year.data == 3
    assert form.dispatch_tax_forms_older.data == 4
    assert form.dispatch_single_documents.data == 5
    assert form.dispatch_note.data == 'Note on dispatch'
    assert form.dispatch_cantonal_tax_office.data == 6
    assert form.dispatch_cantonal_scan_center.data == 7
    assert form.return_date.data == date(2019, 1, 10)
    assert form.return_boxes.data == 8
    assert form.return_tax_forms_current_year.data == 9
    assert form.return_tax_forms_last_year.data == 10
    assert form.return_tax_forms_older.data == 11
    assert form.return_single_documents.data == 12
    assert form.return_unscanned_tax_forms_current_year.data == 13
    assert form.return_unscanned_tax_forms_last_year.data == 14
    assert form.return_unscanned_tax_forms_older.data == 15
    assert form.return_unscanned_single_documents.data == 16
    assert form.return_note.data == 'Note on return'

    form.type.data = 'express'
    form.municipality_id.data = municipality_2.id.hex
    form.dispatch_date.data = date(2019, 1, 6)
    form.dispatch_boxes.data = 10
    form.dispatch_tax_forms_current_year.data = 20
    form.dispatch_tax_forms_last_year.data = 30
    form.dispatch_tax_forms_older.data = 40
    form.dispatch_single_documents.data = 50
    form.dispatch_note.data = 'A note on the dispatch'
    form.dispatch_cantonal_tax_office.data = 60
    form.dispatch_cantonal_scan_center.data = 70
    form.return_date.data = date(2019, 1, 10)
    form.return_boxes.data = 80
    form.return_tax_forms_current_year.data = 90
    form.return_tax_forms_last_year.data = 100
    form.return_tax_forms_older.data = 110
    form.return_single_documents.data = 120
    form.return_unscanned_tax_forms_current_year.data = 130
    form.return_unscanned_tax_forms_last_year.data = 140
    form.return_unscanned_tax_forms_older.data = 150
    form.return_unscanned_single_documents.data = 160
    form.return_note.data = 'A note on the return'

    form.update_model(model)
    assert model.municipality_id == municipality_2.id.hex
    assert model.type == 'express'
    assert model.dispatch_date == date(2019, 1, 6)
    assert model.dispatch_boxes == 10
    assert model.dispatch_tax_forms_current_year == 20
    assert model.dispatch_tax_forms_last_year == 30
    assert model.dispatch_tax_forms_older == 40
    assert model.dispatch_single_documents == 50
    assert model.dispatch_note == 'A note on the dispatch'
    assert model.dispatch_cantonal_tax_office == 60
    assert model.dispatch_cantonal_scan_center == 70
    assert model.return_date == date(2019, 1, 10)
    assert model.return_boxes == 80
    assert model.return_tax_forms_current_year == 90
    assert model.return_tax_forms_last_year == 100
    assert model.return_tax_forms_older == 110
    assert model.return_single_documents == 120
    assert model.return_unscanned_tax_forms_current_year == 130
    assert model.return_unscanned_tax_forms_last_year == 140
    assert model.return_unscanned_tax_forms_older == 150
    assert model.return_unscanned_single_documents == 160
    assert model.return_note == 'A note on the return'

    # Test validation
    with freeze_time("2019-01-05"):
        form = UnrestrictedScanJobForm()
        form.request = Request(session, groupid=municipality_1.id.hex)
        form.on_request()
        assert not form.validate()

        form = UnrestrictedScanJobForm(PostData({
            'municipality_id': municipality_1.id.hex,
            'type': 'normal',
            'dispatch_date': '2019-01-18'
        }))
        form.request = Request(session, groupid=municipality_1.id.hex)
        form.on_request()


def test_daily_list_selection_form(session):
    form = DailyListSelectionForm()
    form.request = Request(session)

    # Test get model
    form.date.data = date(2019, 1, 1)
    form.type.data = 'boxes'
    model = form.get_model()
    assert model.__class__.__name__ == 'DailyListBoxes'
    assert model.date == date(2019, 1, 1)

    # Test get model
    form.date.data = date(2019, 1, 1)
    form.type.data = 'boxes_and_forms'
    model = form.get_model()
    assert model.__class__.__name__ == 'DailyListBoxesAndForms'
    assert model.date == date(2019, 1, 1)


def test_report_selection_form(session):
    municipalities = MunicipalityCollection(session)
    adlikon = municipalities.add(name="Adlikon", bfs_number=21)
    aesch = municipalities.add(name="Aesch", bfs_number=241)

    # Test choices
    form = ReportSelectionForm()
    form.request = Request(session)
    form.on_request()
    assert form.municipality_id.choices == [
        (adlikon.id.hex, 'Adlikon (21)'),
        (aesch.id.hex, 'Aesch (241)')
    ]

    # Test get model
    form.start.data = date(2019, 1, 1)
    form.end.data = date(2019, 1, 31)
    form.scan_job_type.data = 'express'
    form.municipality_id.data = aesch.id.hex

    form.report_type.data = 'boxes'
    model = form.get_model()
    assert model.__class__.__name__ == 'ReportBoxes'
    assert model.start == date(2019, 1, 1)
    assert model.end == date(2019, 1, 31)

    form.report_type.data = 'boxes_and_forms'
    model = form.get_model()
    assert model.__class__.__name__ == 'ReportBoxesAndForms'
    assert model.start == date(2019, 1, 1)
    assert model.end == date(2019, 1, 31)
    assert model.type == 'express'

    form.report_type.data = 'forms'
    model = form.get_model()
    assert model.__class__.__name__ == 'ReportFormsByMunicipality'
    assert model.start == date(2019, 1, 1)
    assert model.end == date(2019, 1, 31)
    assert model.type == 'express'
    assert model.municipality_id == aesch.id.hex

    form.report_type.data = 'delivery'
    model = form.get_model()
    assert model.__class__.__name__ == 'ReportBoxesAndFormsByDelivery'
    assert model.start == date(2019, 1, 1)
    assert model.end == date(2019, 1, 31)
    assert model.type == 'express'
    assert model.municipality_id == aesch.id.hex


def test_notification_form():
    # Test apply / update
    form = NotificationForm()
    notification = Notification(title="Title", text="Text")

    form.apply_model(notification)
    assert form.title.data == "Title"
    assert form.text.data == "Text"

    form.title.data = "Title"
    form.text.data = "Text"
    form.update_model(notification)
    assert notification.title == "Title"
    assert notification.text == "Text"

    # Test validation
    form = NotificationForm()
    assert not form.validate()

    form = NotificationForm(PostData({
        'title': "Title",
        'text': "Text"
    }))
    assert form.validate()


def test_scan_jobs_form(session):
    scan_jobs = ScanJobCollection(session)

    # Test apply
    form = ScanJobsForm()
    form.apply_model(scan_jobs)
    assert form.sort_by.data is None
    assert form.sort_order.data is None
    assert form.from_date.data is None
    assert form.to_date.data is None
    assert form.type.data == ['normal', 'express']
    assert form.term.data is None

    scan_jobs.sort_by = 'delivery_number'
    scan_jobs.sort_order = 'ascending'
    scan_jobs.from_date = date(2010, 1, 1)
    scan_jobs.to_date = date(2010, 12, 31)
    scan_jobs.type = ['express']
    scan_jobs.term = 'term'

    form.apply_model(scan_jobs)

    assert form.sort_by.data == 'delivery_number'
    assert form.sort_order.data == 'ascending'
    assert form.from_date.data == date(2010, 1, 1)
    assert form.to_date.data == date(2010, 12, 31)
    assert form.type.data == ['express']
    assert form.term.data == 'term'

    # Test validation
    form = ScanJobsForm()
    assert form.validate()


def test_unrestricted_scan_jobs_form(session):
    scan_jobs = ScanJobCollection(session)
    municipalities = MunicipalityCollection(session)
    municipalities.add(name="Adlikon", bfs_number=21)
    municipalities.add(name="Aesch", bfs_number=241)

    # Test on request / popluate
    form = UnrestrictedScanJobsForm()
    form.request = Request(session)
    form.on_request()
    assert [m[1] for m in form.municipality_id.choices] == [
        'Adlikon (21)',
        'Aesch (241)'
    ]

    # Test apply
    form.apply_model(scan_jobs)
    assert form.sort_by.data is None
    assert form.sort_order.data is None
    assert form.from_date.data is None
    assert form.to_date.data is None
    assert form.type.data == ['normal', 'express']
    assert form.municipality_id.data is None
    assert form.term.data is None

    scan_jobs.sort_by = 'delivery_number'
    scan_jobs.sort_order = 'ascending'
    scan_jobs.from_date = date(2010, 1, 1)
    scan_jobs.to_date = date(2010, 12, 31)
    scan_jobs.type = ['express']
    scan_jobs.term = 'term'
    scan_jobs.municipality_id = [municipalities.query().first().id]

    form.apply_model(scan_jobs)

    assert form.sort_by.data == 'delivery_number'
    assert form.sort_order.data == 'ascending'
    assert form.from_date.data == date(2010, 1, 1)
    assert form.to_date.data == date(2010, 12, 31)
    assert form.type.data == ['express']
    assert form.term.data == 'term'
    assert form.municipality_id.data == [municipalities.query().first().id]

    # Test validation
    form = UnrestrictedScanJobsForm()
    assert form.validate()


def test_create_invoices_form(session):
    municipalities = MunicipalityCollection(session)
    adlikon = municipalities.add(name="Adlikon", bfs_number=21)
    aesch = municipalities.add(name="Aesch", bfs_number=241)

    # Test choices
    form = CreateInvoicesForm()
    form.request = Request(session)
    form.on_request()
    assert form.vat.choices == [
        ('without_vat', 'Without VAT'),
        ('with_vat', 'With VAT')
    ]
    assert form.municipality_id.choices == [
        ('-', '-'),
        (adlikon.id.hex, 'Adlikon (21)'),
        (aesch.id.hex, 'Aesch (241)')
    ]

    # Test update model
    form.from_date.data = date(2019, 1, 1)
    form.to_date.data = date(2019, 1, 31)
    form.cs2_user.data = '123456'
    form.subject.data = 'Rechnungen'
    form.municipality_id.data = aesch.id.hex
    form.accounting_unit.data = '99999'
    form.revenue_account.data = '987654321'
    form.vat.data = 'with_vat'

    model = Invoice(session)
    form.update_model(model)
    assert model.from_date == date(2019, 1, 1)
    assert model.to_date == date(2019, 1, 31)
    assert model.cs2_user == '123456'
    assert model.subject == 'Rechnungen'
    assert model.municipality_id == aesch.id.hex
    assert model.accounting_unit == '99999'
    assert model.revenue_account == '987654321'
    assert model.vat == 'with_vat'

    form.municipality_id.data = '-'
    form.update_model(model)
    assert model.municipality_id is None



from datetime import date
from onegov.user import User
from onegov.user import UserCollection
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import NotificationCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.layouts import AddMunicipalityLayout
from onegov.wtfs.layouts import AddNotificationLayout
from onegov.wtfs.layouts import AddScanJobLayout
from onegov.wtfs.layouts import AddUserLayout
from onegov.wtfs.layouts import DailyListBoxesAndFormsLayout
from onegov.wtfs.layouts import DailyListBoxesLayout
from onegov.wtfs.layouts import DailyListLayout
from onegov.wtfs.layouts import DefaultLayout
from onegov.wtfs.layouts import DeleteMunicipalityDatesLayout
from onegov.wtfs.layouts import DeliveryNoteLayout
from onegov.wtfs.layouts import EditMunicipalityLayout
from onegov.wtfs.layouts import EditNotificationLayout
from onegov.wtfs.layouts import EditScanJobLayout
from onegov.wtfs.layouts import EditUserLayout
from onegov.wtfs.layouts import ImportMunicipalityDataLayout
from onegov.wtfs.layouts import InvoiceLayout
from onegov.wtfs.layouts import MailLayout
from onegov.wtfs.layouts import MunicipalitiesLayout
from onegov.wtfs.layouts import MunicipalityLayout
from onegov.wtfs.layouts import NotificationLayout
from onegov.wtfs.layouts import NotificationsLayout
from onegov.wtfs.layouts import PaymentTypesLayout
from onegov.wtfs.layouts import ReportBoxesAndFormsByDeliveryLayout
from onegov.wtfs.layouts import ReportBoxesAndFormsLayout
from onegov.wtfs.layouts import ReportBoxesLayout
from onegov.wtfs.layouts import ReportFormsByMunicipalityLayout
from onegov.wtfs.layouts import ReportLayout
from onegov.wtfs.layouts import ScanJobLayout
from onegov.wtfs.layouts import ScanJobsLayout
from onegov.wtfs.layouts import UserLayout
from onegov.wtfs.layouts import UsersLayout
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import Notification
from onegov.wtfs.models import Report
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models import ReportFormsByMunicipality
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import AddModelUnrestricted
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import EditModelUnrestricted
from onegov.wtfs.security import ViewModel


class DummyPrincipal(object):
    pass


class DummyApp(object):
    principal = DummyPrincipal()
    theme_options = {}


class DummyRequest(object):
    app = DummyApp()
    is_logged_in = False
    locale = 'de_CH'
    url = ''

    def __init__(self, session=None, roles=[], includes=[], permissions=[]):
        self.session = session
        self.roles = roles
        self.permissions = permissions
        self.includes = includes

    def has_role(self, *roles):
        return any((role in self.roles for role in roles))

    def has_permission(self, model, permission):
        if self.has_role('admin'):
            return permission in {
                AddModel,
                AddModelUnrestricted,
                DeleteModel,
                EditModel,
                EditModelUnrestricted,
                ViewModel
            }
        if self.has_role('editor'):
            return permission in {
                AddModel,
                DeleteModel,
                EditModel,
                ViewModel
            }
        return permission in self.permissions

    def translate(self, text):
        return str(text)

    def include(self, *args, **kwargs):
        self.includes.extend(args)

    def link(self, model, name=''):
        if isinstance(model, str):
            return f'{model}/{name}'
        return f'{model.__class__.__name__}/{name}'

    def exclude_invisible(self, objects):
        return objects

    def new_csrf_token(self):
        return 'x'


def path(links):
    return '/'.join([link.attrs['href'].strip('/') for link in links])


def hrefs(items):
    for item in items:
        if hasattr(item, 'links'):
            for ln in item.links:
                yield (
                    ln.attrs.get('href')
                    or ln.attrs.get('ic-delete-from')
                    or ln.attrs.get('ic-post-to')
                )
        else:
            yield (
                item.attrs.get('href')
                or item.attrs.get('ic-delete-from')
                or item.attrs.get('ic-post-to')
            )


def test_default_layout(wtfs_app):
    request = DummyRequest()
    request.app = wtfs_app
    request.session = wtfs_app.session()
    model = None

    layout = DefaultLayout(model, request)
    assert layout.title == ""
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'Principal'
    assert layout.static_path == 'Principal/static'
    assert layout.app_version
    assert layout.request.includes == ['frameworks', 'chosen', 'common']
    assert layout.top_navigation == []
    assert layout.cancel_url == ''
    assert layout.success_url == ''
    assert layout.homepage_url == 'Principal/'
    assert layout.login_url == 'Auth/login'
    assert layout.logout_url is None
    assert layout.users_url == 'UserCollection/'
    assert layout.municipalities_url == 'MunicipalityCollection/'
    assert layout.scan_jobs_url == 'ScanJobCollection/'
    assert layout.payment_types_url == 'PaymentTypeCollection/'
    assert layout.current_year == date.today().year

    # Login
    request.is_logged_in = True
    request.roles = ['admin']
    layout = DefaultLayout(model, request)
    assert layout.login_url is None
    assert layout.logout_url == 'Auth/logout'
    assert list(hrefs(layout.top_navigation)) == [
        'ScanJobCollection/',
        'DailyList/',
        'Report/',
        'Invoice/',
        'UserCollection/',
        'MunicipalityCollection/',
        'NotificationCollection/'
    ]


def test_mail_layout():
    request = DummyRequest()
    model = None

    layout = MailLayout(model, request)
    assert layout.primary_color == '#fff'


def test_municipality_layouts():
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])

    # Municipality collection
    model = MunicipalityCollection(None)
    layout = MunicipalitiesLayout(model, request)
    assert layout.title == 'Municipalities'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/MunicipalityCollection'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = MunicipalitiesLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'MunicipalityCollection/import-data',
        'MunicipalityCollection/add'
    ]

    # ... add
    layout = AddMunicipalityLayout(model, request)
    assert layout.title == 'Add municipality'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/MunicipalityCollection/#'
    )
    assert layout.cancel_url == 'MunicipalityCollection/'
    assert layout.success_url == 'MunicipalityCollection/'

    layout = AddMunicipalityLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    # ... import data
    layout = ImportMunicipalityDataLayout(model, request)
    assert layout.title == 'Import data'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/MunicipalityCollection/#'
    )
    assert layout.cancel_url == 'MunicipalityCollection/'
    assert layout.success_url == 'MunicipalityCollection/'

    layout = ImportMunicipalityDataLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    # Municipality
    model = Municipality(name='Boppelsen')
    layout = MunicipalityLayout(model, request)
    assert layout.title == 'Boppelsen'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/MunicipalityCollection/#'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = MunicipalityLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'Municipality/edit',
        'Municipality/delete-dates',
        'Municipality/?csrf-token=x'
    ]

    # ... edit
    layout = EditMunicipalityLayout(model, request)
    assert layout.title == 'Edit municipality'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/MunicipalityCollection/Municipality/#'
    )
    assert layout.cancel_url == 'Municipality/'
    assert layout.success_url == 'MunicipalityCollection/'

    layout = EditMunicipalityLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    # ... delete dates
    layout = DeleteMunicipalityDatesLayout(model, request)
    assert layout.title == 'Delete pick-up dates'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/MunicipalityCollection/Municipality/#'
    )
    assert layout.cancel_url == 'Municipality/'
    assert layout.success_url == 'Municipality/'

    layout = DeleteMunicipalityDatesLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []


def test_user_layouts():
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])
    request_editor = DummyRequest(roles=['editor'])

    # User collection
    model = UserCollection(None)
    layout = UsersLayout(model, request)
    assert layout.title == 'Users'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/UserCollection'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = UsersLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'UserCollection/add-unrestricted'
    ]

    layout = UsersLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == ['UserCollection/add']

    # .. add
    layout = AddUserLayout(model, request)
    assert layout.title == 'Add user'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/UserCollection/#'
    )
    assert layout.cancel_url == 'UserCollection/'
    assert layout.success_url == 'UserCollection/'

    layout = AddUserLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    layout = AddUserLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == []

    # User
    model = User(
        realname='Hans Muster',
        username="hans.muster@winterthur.ch",
        role='member',
        password='1234'
    )
    layout = UserLayout(model, request)
    assert layout.title == 'Hans Muster'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/UserCollection/#'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = UserLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'User/edit-unrestricted',
        'User/?csrf-token=x'
    ]

    layout = UserLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == [
        'User/edit',
        'User/?csrf-token=x'
    ]

    # ... edit
    layout = EditUserLayout(model, request)
    assert layout.title == 'Edit user'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/UserCollection/User/#'
    assert layout.cancel_url == 'User/'
    assert layout.success_url == 'UserCollection/'

    layout = EditUserLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    layout = EditUserLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == []


def test_scan_job_layouts(session):
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])
    request_editor = DummyRequest(roles=['editor'])

    municipalities = MunicipalityCollection(session)
    municipality = municipalities.add(name='Boppelsen', bfs_number=82)

    # ScanJob collection
    model = ScanJobCollection(None)
    layout = ScanJobsLayout(model, request)
    assert layout.title == 'Scan jobs'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/ScanJobCollection'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = ScanJobsLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'ScanJobCollection/add-unrestricted'
    ]

    layout = ScanJobsLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == ['ScanJobCollection/add']

    # .. add
    layout = AddScanJobLayout(model, request)
    assert layout.title == 'Add scan job'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/ScanJobCollection/#'
    assert layout.cancel_url == 'ScanJobCollection/'
    assert layout.success_url == 'ScanJobCollection/'

    layout = AddScanJobLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    layout = AddScanJobLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == []

    # ScanJob
    model = ScanJobCollection(session).add(
        type='normal',
        dispatch_date=date(2019, 1, 1),
        municipality_id=municipality.id
    )
    layout = ScanJobLayout(model, request)
    assert layout.title.interpolate() == 'Scan job no. 1'
    assert list(hrefs(layout.editbar_links)) == ['ScanJob/delivery-note']
    assert path(layout.breadcrumbs) == 'DummyPrincipal/ScanJobCollection/#'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = ScanJobLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'ScanJob/delivery-note',
        'ScanJob/edit-unrestricted',
        'ScanJob/?csrf-token=x'
    ]

    layout = ScanJobLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == [
        'ScanJob/delivery-note',
        'ScanJob/edit',
        'ScanJob/?csrf-token=x'
    ]

    # ... edit
    layout = EditScanJobLayout(model, request)
    assert layout.title == 'Edit scan job'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/ScanJobCollection/ScanJob/#'
    )
    assert layout.cancel_url == 'ScanJob/'
    assert layout.success_url == 'ScanJob/'

    layout = EditScanJobLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    layout = EditScanJobLayout(model, request_editor)
    assert list(hrefs(layout.editbar_links)) == []

    # ... delivery note
    layout = DeliveryNoteLayout(model, request)
    assert layout.title == 'Delivery note'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/ScanJobCollection/ScanJob/#'
    )


def test_daily_list_layouts(session):
    request = DummyRequest()

    model = DailyList()
    layout = DailyListLayout(model, request)
    assert layout.title == 'Daily list'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/DailyList'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = DailyListBoxes(session, date_=date(2019, 1, 1))
    layout = DailyListBoxesLayout(model, request)
    assert layout.title == 'Boxes'
    assert layout.subtitle == 'Dienstag 01. Januar 2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/DailyList/#/DailyListBoxes'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = DailyListBoxesAndForms(session, date_=date(2019, 1, 1))
    layout = DailyListBoxesAndFormsLayout(model, request)
    assert layout.title == 'Boxes and forms'
    assert layout.subtitle == 'Dienstag 01. Januar 2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/DailyList/#/DailyListBoxesAndForms'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''


def test_report_layouts(session):
    municipality = MunicipalityCollection(session).add(
        name="Adlikon",
        bfs_number=21,
    )

    request = DummyRequest()

    model = Report(session)
    layout = ReportLayout(model, request)
    assert layout.title == 'Report'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/Report'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = ReportBoxes(
        session,
        start=date(2019, 1, 1),
        end=date(2019, 1, 31)
    )
    layout = ReportBoxesLayout(model, request)
    assert layout.title == 'Report boxes'
    assert layout.subtitle == '01.01.2019-31.01.2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == 'DummyPrincipal/Report/#/ReportBoxes'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = ReportBoxesAndForms(
        session,
        start=date(2019, 1, 1),
        end=date(2019, 1, 31)
    )
    layout = ReportBoxesAndFormsLayout(model, request)
    assert layout.title == 'Report boxes and forms'
    assert layout.subtitle == '01.01.2019-31.01.2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/Report/#/ReportBoxesAndForms'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = ReportFormsByMunicipality(
        session,
        start=date(2019, 1, 1),
        end=date(2019, 1, 31),
        municipality_id=municipality.id
    )
    layout = ReportFormsByMunicipalityLayout(model, request)
    assert layout.title == 'Report forms'
    assert layout.subtitle == 'Adlikon 01.01.2019-31.01.2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/Report/#/ReportFormsByMunicipality'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    model = ReportBoxesAndFormsByDelivery(
        session,
        start=date(2019, 1, 1),
        end=date(2019, 1, 31),
        type='normal',
        municipality_id=municipality.id
    )
    layout = ReportBoxesAndFormsByDeliveryLayout(model, request)
    assert layout.title == 'Report boxes and forms by delivery'
    assert layout.subtitle == 'Adlikon (21) 01.01.2019-31.01.2019'
    assert list(hrefs(layout.editbar_links)) == ['#']
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/Report/#/ReportBoxesAndFormsByDelivery'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''


def test_notification_layouts():
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])

    # Notification collection
    model = NotificationCollection(None)
    layout = NotificationsLayout(model, request)
    assert layout.title == 'Notifications'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/NotificationCollection'
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = NotificationsLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == ['NotificationCollection/add']

    # .. add
    layout = AddNotificationLayout(model, request)
    assert layout.title == 'Add notification'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/NotificationCollection/#'
    )
    assert layout.cancel_url == 'NotificationCollection/'
    assert layout.success_url == 'NotificationCollection/'

    layout = AddNotificationLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []

    # Notification
    model = Notification(title="Title")
    layout = NotificationLayout(model, request)
    assert layout.title == 'Title'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/NotificationCollection/#'
    )
    assert layout.cancel_url == ''
    assert layout.success_url == ''

    layout = NotificationLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'Notification/edit',
        'Notification/?csrf-token=x'
    ]

    # ... edit
    layout = EditNotificationLayout(model, request)
    assert layout.title == 'Edit notification'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == (
        'DummyPrincipal/NotificationCollection/Notification/#'
    )
    assert layout.cancel_url == 'Notification/'
    assert layout.success_url == 'Notification/'

    layout = EditNotificationLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == []


def test_invoice_layouts(session):
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])

    model = Invoice(session)
    layout = InvoiceLayout(model, request)
    assert layout.title == 'Create invoice'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/Invoice'
    assert layout.cancel_url == 'Invoice/'
    assert layout.success_url == 'Invoice/'

    layout = InvoiceLayout(model, request_admin)
    assert list(hrefs(layout.editbar_links)) == [
        'PaymentTypeCollection/',
    ]


def test_payments_type_layouts(session):
    request = DummyRequest()
    request_admin = DummyRequest(roles=['admin'])

    model = PaymentTypeCollection(session)
    layout = PaymentTypesLayout(model, request)
    assert layout.title == 'Manage payment types'
    assert layout.editbar_links == []
    assert path(layout.breadcrumbs) == 'DummyPrincipal/PaymentTypeCollection'
    assert layout.cancel_url == 'Invoice/'
    assert layout.success_url == 'Invoice/'

    layout = PaymentTypesLayout(model, request_admin)
    assert layout.editbar_links == []



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text


class PaymentType(Base, TimestampMixin):
    """ Payment types. """

    __tablename__ = 'wtfs_payment_type'

    #: The name of the type
    name = Column(Text, primary_key=True)

    #: The price per quantity times hundred, used for invoices. Typically
    #: 700 (normal) or 850 (special).
    _price_per_quantity = Column('price_per_quantity', Integer, nullable=False)

    @property
    def price_per_quantity(self):
        return (self._price_per_quantity or 0) / 100

    @price_per_quantity.setter
    def price_per_quantity(self, value):
        self._price_per_quantity = (value or 0) * 100



from onegov.chat import Message
from onegov.core.orm.mixins import meta_property


class Notification(Message):
    """ A changelog entry for an official notice. """

    __mapper_args__ = {'polymorphic_identity': 'wtfs_notification'}

    title = meta_property('title')

    @classmethod
    def create(cls, request, title='', text=''):

        return cls.bound_messages(request.session).add(
            channel_id=request.identity.application_id,
            owner=request.identity.userid,
            text=text,
            meta={'title': title}
        )



from onegov.core.security import Public
from onegov.user import User
from onegov.user import UserCollection
from onegov.wtfs import WtfsApp
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import Notification
from onegov.wtfs.models import ScanJob


class AddModel(object):
    """ The permission to add a given model. """


class AddModelUnrestricted(object):
    """ The permission to add given model without any restrictions. """


class EditModel(object):
    """ The permission to edit a given model. """


class EditModelUnrestricted(object):
    """ The permission to edit a given model without any restrictions. """


class DeleteModel(object):
    """ The permission to delete a given model. """


class ViewModel(object):
    """ The permission to view a given model. """


class ViewModelUnrestricted(object):
    """ The permission to view a given model without any restrictions. """


def same_group(model, identity):
    """ Returns True, if the given model is in the same user group/municipality
    as the given identy.

    """
    if hasattr(model, 'group_id'):
        if model.group_id and identity.groupid:
            return model.group_id.hex == identity.groupid
    elif hasattr(model, 'municipality_id'):
        if model.municipality_id and identity.groupid:
            return model.municipality_id.hex == identity.groupid
    return False


@WtfsApp.setting_section(section="roles")
def get_roles_setting():
    return {
        'admin': set((
            AddModel,
            AddModelUnrestricted,
            EditModel,
            EditModelUnrestricted,
            DeleteModel,
            ViewModel,
            ViewModelUnrestricted,
            Public,
        )),
        'editor': set((
            Public,
        )),
        'member': set((
            Public,
        )),
        'anonymous': set((
            Public,
        )),
    }


@WtfsApp.permission_rule(model=Municipality, permission=object)
def has_permission_municipality(app, identity, model, permission):
    # Municipalities with data and/or users cannot not be deleted
    if permission in {DeleteModel}:
        if model.users.first():
            return False
        if model.has_data:
            return False

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=UserCollection, permission=object)
def has_permission_users(app, identity, model, permission):
    # Editors may view and add users
    if identity.role == 'editor':
        if permission in {ViewModel, AddModel}:
            return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=User, permission=object)
def has_permission_user(app, identity, model, permission):
    # Admins can not be managed through the web
    if model.role == 'admin':
        if permission not in {Public}:
            return False

    # Editors may view, edit and delete members within the same group
    if identity.role == 'editor':
        if permission in {ViewModel, EditModel, DeleteModel}:
            if model.role == 'member':
                if same_group(model, identity):
                    return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=ScanJobCollection, permission=object)
def has_permission_scan_jobs(app, identity, model, permission):
    # Editors and members of groups may view and add scan jobs
    if identity.role in ('editor', 'member'):
        if identity.groupid:
            if permission in {ViewModel, AddModel}:
                return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=ScanJob, permission=object)
def has_permission_scan_job(app, identity, model, permission):
    # Editors and members of groups may view and edit scan jobs within
    # the same group
    if identity.role in ('editor', 'member'):
        if identity.groupid:
            if permission in {ViewModel, EditModel}:
                if same_group(model, identity):
                    return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=DailyList, permission=object)
def has_permission_daily_list(app, identity, model, permission):
    # Members without groups (transport company) may view the daily list
    # selection form
    if identity.role == 'member':
        if not identity.groupid:
            if permission in {ViewModel}:
                return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=DailyListBoxes, permission=object)
def has_permission_daily_list_boxes(app, identity, model, permission):
    # Members without groups (transport company) may view the daily list boxes
    if identity.role == 'member':
        if not identity.groupid:
            if permission in {ViewModel}:
                return True

    return permission in getattr(app.settings.roles, identity.role)


@WtfsApp.permission_rule(model=Notification, permission=object)
def has_permission_notificastion(app, identity, model, permission):
    # Everybody may view notifications
    if permission in {ViewModel}:
        return True

    return permission in getattr(app.settings.roles, identity.role)



from onegov.core.collection import GenericCollection
from onegov.wtfs.models import PaymentType


class PaymentTypeCollection(GenericCollection):

    @property
    def model_class(self):
        return PaymentType



from morepath import redirect
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.collections import NotificationCollection
from onegov.wtfs.forms import NotificationForm
from onegov.wtfs.layouts import AddNotificationLayout
from onegov.wtfs.layouts import EditNotificationLayout
from onegov.wtfs.layouts import NotificationLayout
from onegov.wtfs.layouts import NotificationsLayout
from onegov.wtfs.models import Notification
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import ViewModel


@WtfsApp.html(
    model=NotificationCollection,
    template='notifications.pt',
    permission=ViewModel
)
def view_notifications(self, request):
    """ View the list of notifications. """

    return {'layout': NotificationsLayout(self, request)}


@WtfsApp.form(
    model=NotificationCollection,
    name='add',
    template='form.pt',
    permission=AddModel,
    form=NotificationForm
)
def add_notification(self, request, form):
    """ Create a new notification. """

    layout = AddNotificationLayout(self, request)

    if form.submitted(request):
        notification = Notification.create(request)
        form.update_model(notification)
        request.message(_("Notification added."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.html(
    model=Notification,
    template='notification.pt',
    permission=ViewModel
)
def view_notification(self, request):
    """ View a single notification. """

    return {'layout': NotificationLayout(self, request)}


@WtfsApp.form(
    model=Notification,
    name='edit',
    template='form.pt',
    permission=EditModel,
    form=NotificationForm
)
def edit_notification(self, request, form):
    """ Edit a notification. """

    layout = EditNotificationLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Notification modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.view(
    model=Notification,
    request_method='DELETE',
    permission=DeleteModel
)
def delete_notification(self, request):
    """ Delete a notification. """

    request.assert_valid_csrf_token()
    NotificationCollection(request.session).delete(self)
    request.message(_("Notification deleted."), 'success')



from morepath import redirect
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import DailyListSelectionForm
from onegov.wtfs.layouts import DailyListBoxesLayout
from onegov.wtfs.layouts import DailyListBoxesAndFormsLayout
from onegov.wtfs.layouts import DailyListLayout
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from onegov.wtfs.security import ViewModel


@WtfsApp.form(
    model=DailyList,
    template='form.pt',
    permission=ViewModel,
    form=DailyListSelectionForm
)
def view_select_report(self, request, form):
    layout = DailyListLayout(self, request)

    if form.submitted(request):
        return redirect(request.link(form.get_model()))

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Show")
    }


@WtfsApp.html(
    model=DailyListBoxes,
    template='daily_list_boxes.pt',
    permission=ViewModel
)
def view_daily_list_boxes(self, request):
    return {'layout': DailyListBoxesLayout(self, request)}


@WtfsApp.html(
    model=DailyListBoxesAndForms,
    template='daily_list_boxes_and_forms.pt',
    permission=ViewModel
)
def view_daily_list_boxes_and_forms(self, request):
    return {'layout': DailyListBoxesAndFormsLayout(self, request)}



from onegov.wtfs.models import Municipality
from onegov.wtfs.models import PaymentType
from transaction import commit


def test_app_principal(wtfs_app):
    assert wtfs_app.principal


def test_app_initial_content(wtfs_app):
    session = wtfs_app.session()
    session.query(PaymentType).delete()

    wtfs_app.add_initial_content()
    commit()

    item = session.query(Municipality).filter_by(name="Andelfingen").first()
    assert item.name == "Andelfingen"
    assert item.bfs_number == 30

    payment_types = dict(
        session.query(PaymentType.name, PaymentType._price_per_quantity)
    )
    assert payment_types == {'normal': 700, 'spezial': 850}



from freezegun import freeze_time
from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_report(client):
    # Add a municipality with dates
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '1'
    add.form['payment_type'] = 'normal'
    assert "Adlikon" in add.form.submit().follow()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n1,5.1.2019".encode('utf-8'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add a scan job
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs/unrestricted').click(href='/add')
        add.form['type'].select("normal")
        add.form['municipality_id'].select(text="Adlikon (1)")
        add.form['dispatch_date'] = "2019-01-05"
        add.form['dispatch_boxes'] = "1111"
        add.form['dispatch_cantonal_tax_office'] = "2222"
        add.form['dispatch_cantonal_scan_center'] = "3333"
        add.form['dispatch_tax_forms_older'] = "4444"
        add.form['dispatch_tax_forms_last_year'] = "5555"
        add.form['dispatch_tax_forms_current_year'] = "6666"
        add.form['dispatch_single_documents'] = "7777"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().maybe_follow()

        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        edit.form['return_date'] = "2019-01-10"
        edit.form['return_boxes'] = "8888"
        edit.form['return_tax_forms_older'] = "9600"
        edit.form['return_tax_forms_last_year'] = "9700"
        edit.form['return_tax_forms_current_year'] = "9800"
        edit.form['return_single_documents'] = "9900"
        edit.form['return_unscanned_tax_forms_older'] = "96"
        edit.form['return_unscanned_tax_forms_last_year'] = "97"
        edit.form['return_unscanned_tax_forms_current_year'] = "98"
        edit.form['return_unscanned_single_documents'] = "99"
        assert "Scan-Auftrag geändert." in edit.form.submit().maybe_follow()

    def get_report(report_type, start, end, scan_job_type='all'):
        select = client.get('/report')
        select.form['start'] = start
        select.form['end'] = end
        select.form['report_type'].select(report_type)
        select.form['scan_job_type'].select(scan_job_type)
        select.form['municipality_id'].select(text='Adlikon (1)')
        return select.form.submit().follow()

    # Boxes
    view = get_report('boxes', '2019-01-01', '2019-01-05')
    assert "1111" in view
    assert "2222" in view
    assert "3333" in view
    assert "8888" in view

    view = get_report('boxes', '2019-01-06', '2019-01-10')
    assert "1111" not in view
    assert "2222" not in view
    assert "3333" not in view
    assert "8888" not in view

    # Boxes and forms
    view = get_report('boxes_and_forms', '2019-01-01', '2019-01-05')
    assert "8888" in view
    assert "9504" in view
    assert "9603" in view
    assert "9702" in view
    assert "9801" in view

    view = get_report('boxes_and_forms', '2019-01-06', '2019-01-10')
    assert "8888" not in view
    assert "9504" not in view
    assert "9603" not in view
    assert "9702" not in view
    assert "9801" not in view

    # Forms
    view = get_report('forms', '2019-01-01', '2019-01-05')
    assert "9504" in view
    assert "9603" in view
    assert "9702" in view

    view = get_report('forms', '2019-01-06', '2019-01-10')
    assert "9504" not in view
    assert "9603" not in view
    assert "9702" not in view

    # By delivery
    view = get_report('delivery', '2019-01-01', '2019-01-05')
    assert "8888" in view
    assert "9504" in view
    assert "9603" in view
    assert "9702" in view
    assert "9801" in view

    view = get_report('delivery', '2019-01-06', '2019-01-10')
    assert "8888" not in view
    assert "9504" not in view
    assert "9603" not in view
    assert "9702" not in view
    assert "9801" not in view


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_report_permissions(mock_method, client):
    client.login_admin()
    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '1'
    add.form['payment_type'] = 'normal'
    added = add.form.submit().follow()
    assert "Adlikon" in added
    id = added.click("Adlikon").request.url.split('/')[-1]
    client.logout()

    urls = (
        '/report/boxes/2019-02-19/2019-02-19',
        '/report/boxes-and-forms/2019-02-19/2019-02-19/all',
        f'/report/forms/2019-02-19/2019-02-19/all/{id}',
    )

    for url in urls:
        client.get(url, status=403)

    client.login_member()
    for url in urls:
        client.get(url, status=403)
    client.logout()

    client.login_editor()
    for url in urls:
        client.get(url, status=403)
    client.logout()

    client.login_admin()
    for url in urls:
        client.get(url)
    client.logout()



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.wtfs import _
from onegov.wtfs.layouts.default import DefaultLayout


class PaymentTypesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Manage payment types")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.request.link(self.model))
        ]

    @cached_property
    def cancel_url(self):
        return self.invoices_url

    @cached_property
    def success_url(self):
        return self.invoices_url



from cached_property import cached_property
from onegov.core.elements import Confirm
from onegov.core.elements import Intercooler
from onegov.core.elements import Link
from onegov.core.elements import LinkGroup
from onegov.wtfs import _
from onegov.wtfs.layouts.default import DefaultLayout
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel


class NotificationsLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Notifications")

    @cached_property
    def editbar_links(self):
        result = []
        if self.request.has_permission(self.model, AddModel):
            result.append(
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Notification"),
                            url=self.request.link(
                                self.model,
                                name='add'
                            ),
                            attrs={'class': 'notification-icon'}
                        )
                    ]
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.notifications_url)
        ]


class NotificationLayout(DefaultLayout):

    @cached_property
    def title(self):
        return self.model.title

    @cached_property
    def editbar_links(self):
        result = []
        if self.request.has_permission(self.model, EditModel):
            result.append(
                Link(
                    text=_("Edit"),
                    url=self.request.link(self.model, 'edit'),
                    attrs={'class': 'edit-icon'}
                )
            )
        if self.request.has_permission(self.model, DeleteModel):
            result.append(
                Link(
                    text=_("Delete"),
                    url=self.csrf_protected_url(
                        self.request.link(self.model)
                    ),
                    attrs={'class': 'delete-icon'},
                    traits=(
                        Confirm(
                            _(
                                "Do you really want to delete this "
                                "notification?"
                            ),
                            _("This cannot be undone."),
                            _("Delete"),
                            _("Cancel")
                        ),
                        Intercooler(
                            request_method='DELETE',
                            redirect_after=self.notifications_url
                        )
                    )
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Notifications"), self.notifications_url),
            Link(self.title, '#')
        ]


class AddNotificationLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Add notification")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Notifications"), self.notifications_url),
            Link(_("Add"), '#')
        ]

    @cached_property
    def cancel_url(self):
        return self.notifications_url

    @cached_property
    def success_url(self):
        return self.notifications_url


class EditNotificationLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Edit notification")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Notifications"), self.notifications_url),
            Link(self.model.title, self.request.link(self.model)),
            Link(_("Edit"), '#')
        ]

    @cached_property
    def cancel_url(self):
        return self.request.link(self.model)

    @cached_property
    def success_url(self):
        return self.request.link(self.model)



from freezegun import freeze_time
from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_daily_job(client):
    # Add a municipality with dates
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "My Municipality"
    add.form['bfs_number'] = '1'
    add.form['payment_type'] = 'normal'
    assert "My Municipality" in add.form.submit().follow()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n1,5.1.2019".encode('utf-8'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add a scan job
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs/unrestricted').click(href='/add')
        add.form['type'].select("normal")
        add.form['municipality_id'].select(text="My Municipality (1)")
        add.form['dispatch_date'] = "2019-01-05"
        add.form['dispatch_boxes'] = "1111"
        add.form['dispatch_cantonal_tax_office'] = "2222"
        add.form['dispatch_cantonal_scan_center'] = "3333"
        add.form['dispatch_tax_forms_older'] = "4444"
        add.form['dispatch_tax_forms_last_year'] = "5555"
        add.form['dispatch_tax_forms_current_year'] = "6666"
        add.form['dispatch_single_documents'] = "7777"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().maybe_follow()

        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        edit.form['return_date'] = "2019-01-10"
        edit.form['return_boxes'] = "8888"
        assert "Scan-Auftrag geändert." in edit.form.submit().maybe_follow()

    def get_daily_list(type, date_):
        select = client.get('/daily-list')
        select.form['type'].select(type)
        select.form['date'] = date_
        return select.form.submit().follow()

    # View daily list boxes
    view = get_daily_list('boxes', '2019-01-05')
    assert "1111" in view
    assert "2222" in view
    assert "3333" in view
    assert "8888" not in view

    view = get_daily_list('boxes', '2019-01-10')
    assert "1111" not in view
    assert "2222" not in view
    assert "3333" not in view
    assert "8888" in view

    # View daily list boxes and forms
    view = get_daily_list('boxes_and_forms', '2019-01-05')
    assert "1111" in view
    assert "2222" in view
    assert "3333" in view
    assert "4444" in view
    assert "5555" in view
    assert "6666" in view

    view = get_daily_list('boxes_and_forms', '2019-01-10')
    assert "1111" not in view
    assert "2222" not in view
    assert "3333" not in view
    assert "4444" not in view
    assert "5555" not in view
    assert "6666" not in view


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_daily_jobs_permissions(mock_method, client):
    client.get(f'/daily-list/boxes/2019-01-10', status=403)
    client.get(f'/daily-list/boxes-and-forms/2019-01-10', status=403)

    client.login_optimo()
    client.get(f'/daily-list/boxes/2019-01-10')
    client.get(f'/daily-list/boxes-and-forms/2019-01-10', status=403)
    client.logout()

    client.login_member()
    client.get(f'/daily-list/boxes/2019-01-10', status=403)
    client.get(f'/daily-list/boxes-and-forms/2019-01-10', status=403)
    client.logout()

    client.login_editor()
    client.get(f'/daily-list/boxes/2019-01-10', status=403)
    client.get(f'/daily-list/boxes-and-forms/2019-01-10', status=403)
    client.logout()

    client.login_admin()
    client.get(f'/daily-list/boxes/2019-01-10')
    client.get(f'/daily-list/boxes-and-forms/2019-01-10')
    client.logout()



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task


@upgrade_task('Add payment types')
def add_payment_types(context):
    session = context.session

    if context.has_table('wtfs_payment_type'):
        query = session.execute('SELECT count(*) FROM wtfs_payment_type')
        if not query.scalar():
            session.execute("""
                INSERT INTO wtfs_payment_type ("name", "price_per_quantity")
                VALUES ('normal', 700), ('spezial', 850);
            """)

            query = session.execute("""
                UPDATE groups
                SET meta = CASE
                    WHEN meta @> '{"_price_per_quantity"\\:850}'::jsonb
                    THEN jsonb_set(meta, '{payment_type}', '"spezial"')
                    ELSE jsonb_set(meta, '{payment_type}', '"normal"')
                END
                WHERE groups.meta ? '_price_per_quantity';
            """)



from cgi import FieldStorage
from datetime import date
from io import BytesIO
from onegov.core.utils import Bunch
from onegov.form import Form
from onegov.wtfs.fields import CsvUploadField
from onegov.wtfs.fields import HintField
from onegov.wtfs.fields import MunicipalityDataUploadField


class PostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


def test_csv_field():
    form = Form()

    field = CsvUploadField()
    field = field.bind(form, 'upload')
    assert field()
    assert len(field.validators) == 2

    def process(content, **kwargs):
        field = CsvUploadField(**kwargs)
        field = field.bind(form, 'upload')

        field_storage = FieldStorage()
        field_storage.file = BytesIO(content)
        field_storage.type = 'text/plain'
        field_storage.filename = 'test.csv'

        field.process(PostData({'upload': field_storage}))
        return field

    # Invalid file
    field = process(b'a,b\n1')
    assert not field.validate(form)
    assert "Not a valid CSV file." in field.errors

    field = process(b'\xc3\x01')
    assert not field.validate(form)
    assert "Not a valid CSV file." in field.errors

    field = process(b'')
    assert not field.validate(form)
    assert "The file is empty." in field.errors

    field = process(b'a,b\n\n1,2')
    assert not field.validate(form)
    assert "The file contains an empty line." in field.errors

    # Invalid headers
    field = process(b'a,b,b\n1,2,2')
    assert not field.validate(form)
    assert "Some column names appear twice." in field.errors

    field = process(b'a,b,b\n1,2,2', rename_duplicate_column_names=True)
    assert field.validate(form)

    field = process(b'a,b\n1,2', expected_headers=['a', 'c'])
    assert not field.validate(form)
    assert "Some columns are missing." in field.errors

    # Valid
    field = process(b'a,b,c\n1,2,3\n4,5,6')
    assert field.validate(form)
    assert [(row.a, row.b, row.c) for row in field.data.lines] == [
        ('1', '2', '3'), ('4', '5', '6')
    ]


def test_municipality_data_upload_field():
    form = Form()

    def process(content, **kwargs):
        field = MunicipalityDataUploadField(**kwargs)
        field = field.bind(form, 'upload')

        field_storage = FieldStorage()
        field_storage.file = BytesIO(content)
        field_storage.type = 'text/plain'
        field_storage.filename = 'test.csv'

        field.process(PostData({'upload': field_storage}))
        return field

    # Invalid
    field = process(
        b'Gemeinde-Nr,Vordefinierte Termine\n'
        b'Adlikon,\n'
    )
    assert not field.validate(form)
    errors = [error.interpolate() for error in field.errors]
    assert "Some rows contain invalid values: 2." in errors

    field = process(
        b'Gemeinde-Nr,Vordefinierte Termine\n'
        b'21,Test\n'
    )
    assert not field.validate(form)
    errors = [error.interpolate() for error in field.errors]
    assert "Some rows contain invalid values: 2." in errors

    # Valid
    field = process(
        b'Gemeinde-Nr,Vordefinierte Termine\n'
        b'21,\n'
    )
    assert field.validate(form)
    assert field.data == {21: {'dates': []}}

    field = process(
        b'Gemeinde-Nr,Vordefinierte Termine,Vordefinierte Termine\n'
        b'21,01.01.2019,07.01.2019\n'
    )
    assert field.validate(form)
    assert field.data == {
        21: {'dates': [date(2019, 1, 1), date(2019, 1, 7)]}
    }


def test_hint_field(wtfs_app):
    def get_translate(for_chameleon):
        return wtfs_app.chameleon_translations.get('de_CH')

    form = Form()
    field = HintField(macro='express_shipment_hint')
    field = field.bind(form, 'hint')
    field.meta.request = Bunch(app=wtfs_app, get_translate=get_translate)

    assert field.validate(form)
    assert "Für dringende Scan-Aufträge" in field()



from onegov_testing import Client as BaseClient
from onegov_testing.utils import create_app
from onegov.core.crypto import hash_password
from onegov.user import User
from onegov.wtfs import WtfsApp
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import PaymentType
from onegov.wtfs.models import Principal
from pytest import fixture
from transaction import commit
from uuid import uuid4


class Client(BaseClient):
    use_intercooler = True

    def login_member(self, to=None):
        return self.login('member@example.org', 'hunter2', to)

    def login_optimo(self, to=None):
        return self.login('optimo@example.org', 'hunter2', to)


def create_wtfs_app(request, temporary_path):
    app = create_app(
        WtfsApp,
        request,
        use_smtp=True,
        depot_backend='depot.io.local.LocalFileStorage',
        depot_storage_path=str(temporary_path),
    )
    app.session_manager.set_locale('de_CH', 'de_CH')

    session = app.session()
    group_id = uuid4()
    session.add(PaymentType(name='normal', _price_per_quantity=700))
    session.add(PaymentType(name='spezial', _price_per_quantity=850))
    session.add(Municipality(
        id=group_id,
        name='My Municipality',
        bfs_number=1,
        payment_type='normal'
    ))
    session.add(User(
        realname='Admin',
        username='admin@example.org',
        password_hash=request.getfixturevalue('wtfs_password'),
        role='admin'
    ))
    session.add(User(
        realname='Editor',
        username='editor@example.org',
        password_hash=request.getfixturevalue('wtfs_password'),
        role='editor',
        group_id=group_id
    ))
    session.add(User(
        realname='Member',
        username='member@example.org',
        password_hash=request.getfixturevalue('wtfs_password'),
        role='member',
        group_id=group_id
    ))
    session.add(User(
        realname='Optimo',
        username='optimo@example.org',
        password_hash=request.getfixturevalue('wtfs_password'),
        role='member'
    ))

    commit()
    return app


@fixture(scope='session')
def wtfs_password():
    return hash_password('hunter2')


@fixture(scope="function")
def principal():
    yield Principal()


@fixture(scope="function")
def wtfs_app(request, temporary_path):
    app = create_wtfs_app(request, temporary_path)
    yield app
    app.session_manager.dispose()


@fixture(scope='function')
def client(wtfs_app):
    return Client(wtfs_app)



from freezegun import freeze_time
from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_scan_job(client):
    # Add municipality dates
    client.login_admin()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n1,5.1.2019".encode('utf-8'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add a scan job
    client.logout()
    client.login_member()
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs').click(href='/add')
        add.form['type'].select("normal")
        add.form['dispatch_date_normal'].select("2019-01-05")
        add.form['dispatch_boxes'] = "1"
        add.form['dispatch_tax_forms_current_year'] = "2"
        add.form['dispatch_tax_forms_last_year'] = "3"
        add.form['dispatch_tax_forms_older'] = "4"
        add.form['dispatch_single_documents'] = "5"
        add.form['dispatch_cantonal_tax_office'] = "6"
        add.form['dispatch_cantonal_scan_center'] = "7"
        add.form['dispatch_note'] = "Bemerkung zur Abholung"
        added = add.form.submit().follow()
        assert "Scan-Auftrag hinzugefügt." in added
        assert "05.01.2019" in added

        message = client.app.smtp.outbox.pop()
        assert message['From'] == 'mails@govikon.ch'
        assert message['To'] == 'member@example.org'
        assert message['Reply-To'] == 'mails@govikon.ch'
        payload = message.get_payload(1).get_payload(decode=True)
        payload = payload.decode('utf-8')
        assert "am 05.01.2019 Ihre Sendung abholen" in payload

    # View scan job
    view = client.get('/scan-jobs').click("05.01.2019")
    assert "Scan-Auftrag Nr. 1" in view
    assert "My Municipality" in view
    assert "05.01.2019" in view
    assert all([f">{number}<" in view for number in range(1, 7 + 1)])
    assert "Lieferung an das Steueramt Winterthur am 05.01.2019" in view
    assert "Bemerkung zur Abholung" in view

    # Delivery note
    view = view.click("Lieferschein")
    assert "My Municipality" in view
    assert "05.01.2019" in view
    assert all([f">{number}<" in view for number in range(1, 5 + 1)])
    assert "Lieferung an das Steueramt Winterthur" in view

    # Edit scan job
    client.logout()
    client.login_editor()
    edit = client.get('/scan-jobs').click("05.01.2019").click("Bearbeiten")
    edit.form['dispatch_note'] = "Eine Bemerkung zur Abholung"
    assert "Scan-Auftrag geändert." in edit.form.submit().follow()

    view = client.get('/scan-jobs').click("05.01.2019")
    assert "Scan-Auftrag Nr. 1" in view
    assert "My Municipality" in view
    assert "05.01.2019" in view
    assert all([f">{number}<" in view for number in range(1, 7 + 1)])
    assert "Lieferung an das Steueramt Winterthur am 05.01.2019" in view
    assert "Eine Bemerkung zur Abholung" in view

    client.logout()
    client.login_admin()

    with freeze_time("2019-01-02"):
        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        edit.form['return_date'] = "2019-01-10"
        edit.form['return_boxes'] = "8"
        edit.form['return_tax_forms_current_year'] = "9"
        edit.form['return_tax_forms_last_year'] = "10"
        edit.form['return_tax_forms_older'] = "11"
        edit.form['return_single_documents'] = "12"
        edit.form['return_unscanned_tax_forms_current_year'] = "13"
        edit.form['return_unscanned_tax_forms_last_year'] = "14"
        edit.form['return_unscanned_tax_forms_older'] = "15"
        edit.form['return_unscanned_single_documents'] = "16"
        edit.form['return_note'] = "Bemerkung zur Rücksendung"
        assert "Scan-Auftrag geändert." in edit.form.submit().follow()

    view = client.get('/scan-jobs/unrestricted').click("05.01.2019")
    assert "Scan-Auftrag Nr. 1" in view
    assert "My Municipality" in view
    assert "05.01.2019" in view
    assert all([f">{number}<" in view for number in range(1, 16 + 1)])
    assert "Lieferung an das Steueramt Winterthur am 05.01.2019" in view
    assert "Rücksendung an My Municipality am 10.01.2019" in view
    assert "Bemerkung zur Abholung" in view
    assert "Bemerkung zur Rücksendung" in view

    # Check the date hints
    with freeze_time("2019-01-01"):
        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        dates = client.post(
            '/dispatch-dates',
            {
                'csrf_token': edit.form['csrf_token'].value,
                'municipality_id': edit.form['municipality_id'].value
            }
        )
        assert "05.01.2019" in dates

    # Delete scan job
    deleted = client.get('/scan-jobs/unrestricted')\
        .click("05.01.2019").click("Löschen")
    assert deleted.status_int == 200
    assert "05.01.2019" not in client.get('/scan-jobs')


def test_views_scan_jobs_filter(client):
    # Add municipality dates
    client.login_admin()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n1,5.1.2019".encode('utf-8'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add scan jobs
    client.logout()
    client.login_member()
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs').click(href='/add')
        add.form['type'].select('normal')
        add.form['dispatch_date_normal'].select("2019-01-05")
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().follow()

        client.logout()
        client.login_editor()
        add = client.get('/scan-jobs').click(href='/add')
        add.form['type'].select('express')
        add.form['dispatch_date_express'] = "2019-01-04"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().follow()

        add.form['type'].select('express')
        add.form['dispatch_date_express'] = "2019-01-06"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().follow()

    # View scan jobs
    view = client.get('/scan-jobs')
    assert view.pyquery('table.scan-jobs td').text() == (
        '06.01.2019 3 express '
        '05.01.2019 1 normal '
        '04.01.2019 2 express'
    )

    view.form.get('type', index=0).checked = False
    view = view.form.submit()
    assert view.pyquery('table.scan-jobs td').text() == (
        '06.01.2019 3 express '
        '04.01.2019 2 express'
    )

    view = view.click("Lieferscheinnummer")
    assert view.pyquery('table.scan-jobs td').text() == (
        '04.01.2019 2 express '
        '06.01.2019 3 express'
    )

    client.logout()
    client.login_admin()
    view = client.get('/scan-jobs/unrestricted')
    assert view.pyquery('table.scan-jobs td').text() == (
        '06.01.2019 3 express My Municipality '
        '05.01.2019 1 normal My Municipality '
        '04.01.2019 2 express My Municipality'
    )


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_scan_jobs_permissions(mock_method, client):
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "My Municipality"
    add.form['bfs_number'] = '1'
    add.form['payment_type'] = 'normal'
    assert "My Municipality" in add.form.submit().follow()

    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs/unrestricted').click(href='/add')
        add.form['type'].select("express")
        add.form['dispatch_date'] = "2019-01-05"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().maybe_follow()
        id = client.get('/scan-jobs/unrestricted').click("05.01.2019")\
            .request.url.split('/')[-1]

    client.logout()

    client.get('/scan-jobs', status=403)
    client.get('/scan-jobs/add', status=403)
    client.get(f'/scan-job/{id}', status=403)
    client.get(f'/scan-job/{id}/edit', status=403)
    client.get('/scan-jobs/unrestricted', status=403)
    client.get('/scan-jobs/add-unrestricted', status=403)
    client.get(f'/scan-job/{id}/edit-unrestricted', status=403)
    client.delete(f'/scan-job/{id}', status=403)

    client.login_member()
    client.get('/scan-jobs')
    client.get('/scan-jobs/add')
    client.get(f'/scan-job/{id}')
    client.get(f'/scan-job/{id}/edit')
    client.get('/scan-jobs/unrestricted', status=403)
    client.get('/scan-jobs/add-unrestricted', status=403)
    client.get(f'/scan-job/{id}/edit-unrestricted', status=403)
    client.delete(f'/scan-job/{id}', status=403)
    client.logout()

    client.login_editor()
    client.get('/scan-jobs')
    client.get('/scan-jobs/add')
    client.get(f'/scan-job/{id}')
    client.get(f'/scan-job/{id}/edit')
    client.get('/scan-jobs/unrestricted', status=403)
    client.get('/scan-jobs/add-unrestricted', status=403)
    client.get(f'/scan-job/{id}/edit-unrestricted', status=403)
    client.delete(f'/scan-job/{id}', status=403)
    client.logout()

    client.login_admin()
    client.get('/scan-jobs')
    client.get('/scan-jobs/add')
    client.get(f'/scan-job/{id}')
    client.get(f'/scan-job/{id}/edit')
    client.get('/scan-jobs/unrestricted')
    client.get('/scan-jobs/add-unrestricted')
    client.get(f'/scan-job/{id}/edit-unrestricted')
    client.delete(f'/scan-job/{id}')
    client.logout()



from csv import writer
from datetime import datetime
from onegov.wtfs.models.scan_job import ScanJob


class Invoice(object):

    def __init__(self, session):
        self.session = session
        self.from_date = None
        self.to_date = None
        self.cs2_user = None
        self.subject = None
        self.municipality_id = None
        self.accounting_unit = None
        self.revenue_account = None
        self.vat = None

    def export(self, file):
        csv = writer(file)
        csv.writerow((
            'Aktuelles Datum (YYYYMMDD1)',
            'Aktuelles Datum (DD.MM.YYYY)',
            'Download Zeitpunkt (HH.MIN.SEK)',
            'GPN-Nr.',
            'Positionsnummer (Hochgezählt)',
            'GPN-Nr.',
            'Betreff',
            'Positionsnummer (Hochgezählt)',
            'Erstellungsdatum',
            'Erstellungsdatum',
            'Erstellungsdatum',
            'GPN-Nr.',
            'Total StE',
            'Lieferscheinnummer',
            'Preis pro Menge exkl. MWST',
            'Preis pro Menge exkl. MWST',
            'Fixe Preiseiheit (immer „1“)',
            'Erstellungsdatum',
            'Adressszusatz der Gemeinde',
            'CS2 Benutzer',
            'Kostenstelle',
            'Erlöskonto',
        ))

        now = datetime.now()
        current_date_1 = f'{now:%Y%m%d}1'
        current_date = f'{now:%d.%m.%Y}'
        current_time = f'{now:%H.%M.%S}'

        query = self.session.query(ScanJob)
        if self.from_date:
            query = query.filter(ScanJob.dispatch_date >= self.from_date)
        if self.to_date:
            query = query.filter(ScanJob.dispatch_date <= self.to_date)
        if self.municipality_id:
            query = query.filter(
                ScanJob.municipality_id == self.municipality_id
            )
        query = query.order_by(ScanJob.dispatch_date)
        for count, scan_job in enumerate(query, start=1):
            municipality = scan_job.municipality
            gpn_number = municipality.gpn_number
            gpn_number = str(gpn_number) if gpn_number is not None else ''
            address_supplement = municipality.address_supplement or ''
            price_per_quantity = municipality.price_per_quantity
            price_per_quantity = str(int(price_per_quantity * 10000000))
            tax_forms = (
                (scan_job.dispatch_tax_forms or 0)
                - (scan_job.return_unscanned_tax_forms or 0)
            )
            csv.writerow((
                current_date_1,
                current_date,
                current_time,
                gpn_number,
                str(count),
                gpn_number,
                self.subject,
                str(count),
                current_date,
                current_date,
                current_date,
                gpn_number,
                str(tax_forms * 1000),
                str(scan_job.delivery_number),
                price_per_quantity,
                price_per_quantity,
                '1',
                current_date,
                address_supplement,
                self.cs2_user,
                self.accounting_unit,
                self.revenue_account,
            ))



from onegov.core.collection import Pagination
from onegov.core.orm.func import unaccent
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import ScanJob
from sqlalchemy import or_


class ScanJobCollection(Pagination):

    batch_size = 20
    initial_sort_by = 'dispatch_date'
    initial_sort_order = 'descending'
    default_sort_order = 'ascending'

    SORT_BYS = (
        'dispatch_date',
        'delivery_number',
        'municipality_id'
    )
    SORT_ORDERS = ('ascending', 'descending')

    def __init__(
        self,
        session,
        page=None,
        group_id=None,
        from_date=None,
        to_date=None,
        type=None,
        municipality_id=None,
        term=None,
        sort_by=None,
        sort_order=None
    ):
        self.session = session
        self.page = page
        self.group_id = group_id
        self.from_date = from_date
        self.to_date = to_date
        self.type = type
        self.municipality_id = municipality_id
        self.term = term
        self.sort_by = sort_by
        self.sort_order = sort_order

    def add(self, **kwargs):
        """ Adds a new scan job. """

        scan_job = ScanJob(**kwargs)
        self.session.add(scan_job)
        self.session.flush()
        return scan_job

    def subset(self):
        return self.query()

    def __eq__(self, other):
        return (
            (self.page or 0) == (other.page or 0)
            and (self.group_id or None) == (other.group_id or None)
            and (self.from_date or None) == (other.from_date or None)
            and (self.to_date or None) == (other.to_date or None)
            and set(self.type or []) == set(other.type or [])
            and (
                set(self.municipality_id or [])
                == set(other.municipality_id or [])
            )
            and (self.term or None) == (other.term or None)
            and (self.sort_by or None) == (other.sort_by or None)
            and (self.sort_order or None) == (other.sort_order or None)
        )

    def default(self):
        """ Returns the jobs unfiltered and ordered by default. """

        return self.__class__(self.session, group_id=self.group_id)

    @property
    def page_index(self):
        """ The current page. """

        return self.page or 0

    def page_by_index(self, page):
        """ Returns the requested page. """

        return self.__class__(
            self.session,
            page=page,
            group_id=self.group_id,
            from_date=self.from_date,
            to_date=self.to_date,
            type=self.type,
            municipality_id=self.municipality_id,
            term=self.term,
            sort_by=self.sort_by,
            sort_order=self.sort_order
        )

    @property
    def current_sort_by(self):
        """ Returns the currently used sorting key.

        Defaults to a reasonable value.

        """
        if self.sort_by in self.SORT_BYS:
            return self.sort_by

        return self.initial_sort_by

    @property
    def current_sort_order(self):
        """ Returns the currently used sorting order.

        Defaults to a reasonable value.

        """
        if self.sort_by in self.SORT_BYS:
            if self.sort_order in self.SORT_ORDERS:
                return self.sort_order

            if self.sort_by == self.initial_sort_by:
                return self.initial_sort_order

            return self.default_sort_order

        return self.initial_sort_order

    def sort_order_by_key(self, sort_by):
        """ Returns the sort order by key.

        Defaults to 'unsorted'.

        """

        if self.current_sort_by == sort_by:
            return self.current_sort_order
        return 'unsorted'

    def by_order(self, sort_by):
        """ Returns the jobs ordered by the given key. """

        sort_order = self.default_sort_order
        if sort_by == self.current_sort_by:
            if self.current_sort_order == 'ascending':
                sort_order = 'descending'
            else:
                sort_order = 'ascending'

        return self.__class__(
            self.session,
            page=None,
            group_id=self.group_id,
            from_date=self.from_date,
            to_date=self.to_date,
            type=self.type,
            municipality_id=self.municipality_id,
            term=self.term,
            sort_by=sort_by,
            sort_order=sort_order
        )

    @property
    def order_by(self):
        """ Returns an SqlAlchemy expression for ordering queries based
        on the current sorting key and ordering.

        """

        if self.current_sort_by == 'municipality_id':
            result = unaccent(Municipality.name)
        else:
            result = getattr(ScanJob, self.current_sort_by, None)
            if not result:
                raise NotImplementedError()

        if self.current_sort_order == 'descending':
            result = result.desc()

        return result

    @property
    def offset(self):
        """ The current position in the batch. """

        return (self.page or 0) * self.batch_size

    @property
    def previous(self):
        """ The previous page. """

        if (self.page or 0) - 1 >= 0:
            return self.page_by_index((self.page or 0) - 1)

    @property
    def next(self):
        """ The next page. """

        if (self.page or 0) + 1 < self.pages_count:
            return self.page_by_index((self.page or 0) + 1)

    def query(self):
        """ Returns the jobs matching to the current filters and order. """

        query = self.session.query(ScanJob)
        query = query.join(Municipality)

        if self.group_id:
            query = query.filter(ScanJob.municipality_id == self.group_id)
        elif self.municipality_id:
            query = query.filter(
                ScanJob.municipality_id.in_(self.municipality_id)
            )
        if self.from_date:
            query = query.filter(ScanJob.dispatch_date >= self.from_date)
        if self.to_date:
            query = query.filter(ScanJob.dispatch_date <= self.to_date)
        if self.type:
            query = query.filter(ScanJob.type.in_(self.type))
        if self.term:
            query = query.filter(
                or_(*[
                    ScanJob.dispatch_note.ilike(f'%{self.term}%'),
                    ScanJob.return_note.ilike(f'%{self.term}%')
                ])
            )

        query = query.order_by(self.order_by)

        return query

    def by_id(self, id):
        """ Returns the scan job with the given ID. """
        return self.query().filter(ScanJob.id == id).first()

    def delete(self, scan_job):
        """ Deletes the given scan job. """
        self.session.delete(scan_job)
        self.session.flush()



from datetime import date
from freezegun import freeze_time
from io import StringIO
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import Notification
from onegov.wtfs.models import PaymentType
from onegov.wtfs.models import PickupDate
from onegov.wtfs.models import Principal
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models import ReportFormsByMunicipality
from onegov.wtfs.models import ScanJob
from uuid import uuid4


def test_principal():
    principal = Principal()
    assert principal


def test_payment_type(session):
    session.add(PaymentType(name='normal', _price_per_quantity=700))
    session.flush()

    payment_type = session.query(PaymentType).one()
    assert payment_type.name == 'normal'
    assert payment_type._price_per_quantity == 700
    assert payment_type.price_per_quantity == 7.0

    payment_type.price_per_quantity = 8.5
    assert payment_type._price_per_quantity == 850
    assert payment_type.price_per_quantity == 8.5


def test_municipality(session):
    session.add(PaymentType(name='normal', _price_per_quantity=700))
    session.add(PaymentType(name='spezial', _price_per_quantity=850))

    session.add(
        Municipality(
            name='Winterthur',
            bfs_number=230,
            address_supplement='Zusatz',
            gpn_number=1120,
            payment_type='normal'
        )
    )
    session.flush()

    municipality = session.query(Municipality).one()
    assert municipality.name == 'Winterthur'
    assert municipality.bfs_number == 230
    assert municipality.address_supplement == 'Zusatz'
    assert municipality.gpn_number == 1120
    assert municipality.price_per_quantity == 7.0
    assert not municipality.has_data

    municipality.payment_type = 'spezial'
    assert municipality.price_per_quantity == 8.5

    municipality.payment_type = 'invalid'
    assert municipality.price_per_quantity == 0

    # PickupDate
    session.add(
        PickupDate(municipality_id=municipality.id, date=date(2019, 1, 1))
    )
    session.add(
        PickupDate(municipality_id=municipality.id, date=date(2019, 1, 7))
    )
    session.add(
        PickupDate(municipality_id=municipality.id, date=date(2019, 1, 14))
    )
    session.flush()
    session.expire_all()

    assert [d.date for d in municipality.pickup_dates] == [
        date(2019, 1, 1), date(2019, 1, 7), date(2019, 1, 14)
    ]
    assert session.query(PickupDate).first().municipality == municipality
    assert municipality.has_data


def test_scan_job(session):
    session.add(Municipality(name='Winterthur', bfs_number=230))
    session.flush()
    municipality = session.query(Municipality).one()

    session.add(
        ScanJob(
            type='normal',
            municipality_id=municipality.id,
            dispatch_date=date(2019, 1, 1),
            dispatch_note='Dispatch note',
            dispatch_boxes=1,
            dispatch_tax_forms_current_year=2,
            dispatch_tax_forms_last_year=3,
            dispatch_tax_forms_older=4,
            dispatch_single_documents=5,
            dispatch_cantonal_tax_office=6,
            dispatch_cantonal_scan_center=7,
            return_date=date(2019, 2, 2),
            return_note='Return note',
            return_boxes=8,
            return_tax_forms_current_year=9,
            return_tax_forms_last_year=10,
            return_tax_forms_older=11,
            return_single_documents=12,
            return_unscanned_tax_forms_current_year=16,
            return_unscanned_tax_forms_last_year=15,
            return_unscanned_tax_forms_older=14,
            return_unscanned_single_documents=13,
        )
    )
    session.flush()

    scan_job = session.query(ScanJob).one()
    assert scan_job.municipality == municipality
    assert scan_job.delivery_number == 1
    assert scan_job.title.interpolate() == 'Scan job no. 1'
    assert scan_job.type == 'normal'
    assert scan_job.municipality_id == municipality.id
    assert scan_job.dispatch_date == date(2019, 1, 1)
    assert scan_job.dispatch_note == 'Dispatch note'
    assert scan_job.dispatch_boxes == 1
    assert scan_job.dispatch_tax_forms_current_year == 2
    assert scan_job.dispatch_tax_forms_last_year == 3
    assert scan_job.dispatch_tax_forms_older == 4
    assert scan_job.dispatch_tax_forms == 2 + 3 + 4
    assert scan_job.dispatch_single_documents == 5
    assert scan_job.dispatch_cantonal_tax_office == 6
    assert scan_job.dispatch_cantonal_scan_center == 7
    assert scan_job.return_date == date(2019, 2, 2)
    assert scan_job.return_note == 'Return note'
    assert scan_job.return_boxes == 8
    assert scan_job.return_tax_forms_current_year == 9
    assert scan_job.return_tax_forms_last_year == 10
    assert scan_job.return_tax_forms_older == 11
    assert scan_job.return_tax_forms == 9 + 10 + 11
    assert scan_job.return_single_documents == 12
    assert scan_job.return_unscanned_tax_forms_current_year == 16
    assert scan_job.return_unscanned_tax_forms_last_year == 15
    assert scan_job.return_unscanned_tax_forms_older == 14
    assert scan_job.return_unscanned_tax_forms == 16 + 15 + 14
    assert scan_job.return_unscanned_single_documents == 13
    assert scan_job.return_scanned_tax_forms_current_year == 9 - 16
    assert scan_job.return_scanned_tax_forms_last_year == 10 - 15
    assert scan_job.return_scanned_tax_forms_older == 11 - 14
    assert scan_job.return_scanned_tax_forms == 9 + 10 + 11 - 16 - 15 - 14
    assert scan_job.return_scanned_single_documents == 12 - 13

    assert session.query(
        ScanJob.dispatch_boxes,
        ScanJob.dispatch_tax_forms_current_year,
        ScanJob.dispatch_tax_forms_last_year,
        ScanJob.dispatch_tax_forms_older,
        ScanJob.dispatch_tax_forms,
        ScanJob.dispatch_single_documents,
        ScanJob.dispatch_cantonal_tax_office,
        ScanJob.dispatch_cantonal_scan_center,
        ScanJob.return_boxes,
        ScanJob.return_tax_forms_current_year,
        ScanJob.return_tax_forms_last_year,
        ScanJob.return_tax_forms_older,
        ScanJob.return_tax_forms,
        ScanJob.return_single_documents,
        ScanJob.return_unscanned_tax_forms_current_year,
        ScanJob.return_unscanned_tax_forms_last_year,
        ScanJob.return_unscanned_tax_forms_older,
        ScanJob.return_unscanned_tax_forms,
        ScanJob.return_unscanned_single_documents,
        ScanJob.return_scanned_tax_forms_current_year,
        ScanJob.return_scanned_tax_forms_last_year,
        ScanJob.return_scanned_tax_forms_older,
        ScanJob.return_scanned_tax_forms,
        ScanJob.return_scanned_single_documents,
    ).one() == (
        1,
        2,
        3,
        4,
        2 + 3 + 4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        9 + 10 + 11,
        12,
        16,
        15,
        14,
        16 + 15 + 14,
        13,
        9 - 16,
        10 - 15,
        11 - 14,
        9 + 10 + 11 - 16 - 15 - 14,
        12 - 13,
    )

    assert municipality.scan_jobs.one() == scan_job
    assert municipality.has_data


def add_report_data(session):
    data = {
        'Adlikon': {
            'municipality_id': uuid4(),
            'bfs_number': 21,
            'jobs': [
                [date(2019, 1, 1), 1, 2, 3, date(2019, 1, 2), 4],
                [date(2019, 1, 2), 3, 2, 1, date(2019, 1, 3), 1],
            ]
        },
        'Aesch': {
            'municipality_id': uuid4(),
            'bfs_number': 241,
            'jobs': [
                [date(2019, 1, 1), 1, 2, 3, date(2019, 1, 2), 4],
                [date(2019, 1, 3), 0, 10, None, date(2019, 1, 4), None],
            ]
        },
        'Altikon': {
            'municipality_id': uuid4(),
            'bfs_number': 211,
            'jobs': [
                [date(2019, 1, 2), 1, 2, 3, date(2019, 1, 2), 4],
            ]
        },
        'Andelfingen': {
            'municipality_id': uuid4(),
            'bfs_number': 30,
            'jobs': [
                [date(2019, 1, 2), 1, 2, 3, date(2019, 1, 4), 4],
            ]
        },
    }

    for name, values in data.items():
        session.add(Municipality(
            name=name,
            id=values['municipality_id'],
            bfs_number=values['bfs_number']
        ))
        for job in values['jobs']:
            double = [2 * v if isinstance(v, int) else v for v in job]
            session.add(
                ScanJob(
                    type='normal',
                    municipality_id=values['municipality_id'],
                    dispatch_date=job[0],
                    dispatch_boxes=job[1],
                    dispatch_cantonal_tax_office=job[2],
                    dispatch_cantonal_scan_center=job[3],
                    dispatch_tax_forms_older=job[1],
                    dispatch_tax_forms_last_year=job[2],
                    dispatch_tax_forms_current_year=job[3],
                    dispatch_single_documents=job[5],
                    return_tax_forms_older=double[1],
                    return_tax_forms_last_year=double[2],
                    return_tax_forms_current_year=double[3],
                    return_single_documents=double[5],
                    return_unscanned_tax_forms_older=job[1],
                    return_unscanned_tax_forms_last_year=job[2],
                    return_unscanned_tax_forms_current_year=job[3],
                    return_unscanned_single_documents=job[5],
                    return_date=job[4],
                    return_boxes=job[5],
                )
            )
        session.flush()
    session.flush()


def test_daily_list_boxes(session):
    daily_list = DailyListBoxes(session, date_=date.today())
    assert daily_list.query().all() == []
    assert daily_list.total() == (0, 0, 0, 0)

    add_report_data(session)

    daily_list = DailyListBoxes(session, date_=date(2019, 1, 1))
    assert daily_list.query().all() == [
        ('Adlikon', 21, 1, 2, 3, 0),
        ('Aesch', 241, 1, 2, 3, 0)
    ]
    assert daily_list.total() == (2, 4, 6, 0)

    daily_list = DailyListBoxes(session, date_=date(2019, 1, 2))
    assert daily_list.query().all() == [
        ('Adlikon', 21, 3, 2, 1, 4),
        ('Aesch', 241, 0, 0, 0, 4),
        ('Altikon', 211, 1, 2, 3, 4),
        ('Andelfingen', 30, 1, 2, 3, 0)
    ]
    assert daily_list.total() == (5, 6, 7, 12)

    daily_list = DailyListBoxes(session, date_=date(2019, 1, 3))
    assert daily_list.query().all() == [
        ('Adlikon', 21, 0, 0, 0, 1),
        ('Aesch', 241, 0, 10, 0, 0)
    ]
    assert daily_list.total() == (0, 10, 0, 1)

    daily_list = DailyListBoxes(session, date_=date(2019, 1, 4))
    assert daily_list.query().all() == [
        ('Aesch', 241, 0, 0, 0, 0),
        ('Andelfingen', 30, 0, 0, 0, 4)
    ]
    assert daily_list.total() == (0, 0, 0, 4)

    daily_list = DailyListBoxes(session, date_=date(2019, 1, 5))
    assert daily_list.query().all() == []
    assert daily_list.total() == (0, 0, 0, 0)


def test_daily_list_boxes_and_forms(session):
    daily_list = DailyListBoxesAndForms(session, date_=date.today())
    assert daily_list.query().all() == []
    assert daily_list.total() == (0, 0, 0, 0, 0, 0, 0)

    add_report_data(session)

    daily_list = DailyListBoxesAndForms(session, date_=date(2019, 1, 1))
    assert daily_list.query().all() == [
        ('Adlikon', 21, 1, 1, 2, 3, 4, 2, 3),
        ('Aesch', 241, 1, 1, 2, 3, 4, 2, 3)
    ]
    assert daily_list.total() == (2, 2, 4, 6, 8, 4, 6)

    daily_list = DailyListBoxesAndForms(session, date_=date(2019, 1, 2))

    assert daily_list.query().all() == [
        ('Adlikon', 21, 3, 3, 2, 1, 1, 2, 1),
        ('Aesch', 241, 0, 0, 0, 0, 0, 0, 0),
        ('Altikon', 211, 1, 1, 2, 3, 4, 2, 3),
        ('Andelfingen', 30, 1, 1, 2, 3, 4, 2, 3)
    ]
    assert daily_list.total() == (5, 5, 6, 7, 9, 6, 7)

    daily_list = DailyListBoxesAndForms(session, date_=date(2019, 1, 3))
    assert daily_list.query().all() == [
        ('Adlikon', 21, 0, 0, 0, 0, 0, 0, 0),
        ('Aesch', 241, 0, 0, 10, 0, 0, 10, 0)
    ]
    assert daily_list.total() == (0, 0, 10, 0, 0, 10, 0)

    daily_list = DailyListBoxesAndForms(session, date_=date(2019, 1, 4))
    assert daily_list.query().all() == [
        ('Aesch', 241, 0, 0, 0, 0, 0, 0, 0),
        ('Andelfingen', 30, 0, 0, 0, 0, 0, 0, 0)
    ]
    assert daily_list.total() == (0, 0, 0, 0, 0, 0, 0)

    daily_list = DailyListBoxesAndForms(session, date_=date(2019, 1, 5))
    assert daily_list.query().all() == []
    assert daily_list.total() == (0, 0, 0, 0, 0, 0, 0)


def test_report_boxes(session):
    def _report(start, end):
        return ReportBoxes(session, start=start, end=end)

    report = _report(date.today(), date.today())
    assert report.query().all() == []
    assert report.total() == (0, 0, 0, 0)

    add_report_data(session)

    report = _report(date(2019, 1, 1), date(2019, 1, 1))
    assert report.query().all() == [
        ('Adlikon', 21, 1, 2, 3, 4),
        ('Aesch', 241, 1, 2, 3, 4)
    ]
    assert report.total() == (2, 4, 6, 8)

    report = _report(date(2019, 1, 2), date(2019, 1, 3))
    assert report.query().all() == [
        ('Adlikon', 21, 3, 2, 1, 1),
        ('Aesch', 241, 0, 10, 0, 0),
        ('Altikon', 211, 1, 2, 3, 4),
        ('Andelfingen', 30, 1, 2, 3, 4)
    ]
    assert report.total() == (5, 16, 7, 9)

    report = _report(date(2019, 1, 4), date(2019, 1, 5))
    assert report.query().all() == [
        ('Aesch', 241, 0, 0, 0, 0),
        ('Andelfingen', 30, 0, 0, 0, 0)
    ]
    assert report.total() == (0, 0, 0, 0)


def test_report_boxes_and_forms(session):
    def _report(start, end):
        return ReportBoxesAndForms(session, start=start, end=end)

    report = _report(date.today(), date.today())
    assert report.query().all() == []
    assert report.total() == (0, 0, 0, 0, 0, 0)

    add_report_data(session)

    report = _report(date(2019, 1, 1), date(2019, 1, 1))
    assert report.query().all() == [
        ('Adlikon', 21, 1, 2, 3, 6, 4, 4),
        ('Aesch', 241, 1, 2, 3, 6, 4, 4)
    ]
    assert report.total() == (2, 4, 6, 12, 8, 8)

    report = _report(date(2019, 1, 2), date(2019, 1, 3))
    assert report.query().all() == [
        ('Adlikon', 21, 3, 2, 1, 6, 1, 1),
        ('Aesch', 241, 0, 10, 0, 10, 0, 0),
        ('Altikon', 211, 1, 2, 3, 6, 4, 4),
        ('Andelfingen', 30, 1, 2, 3, 6, 4, 4)
    ]
    assert report.total() == (5, 16, 7, 28, 9, 9)

    report = _report(date(2019, 1, 4), date(2019, 1, 5))
    assert report.query().all() == [
        ('Aesch', 241, 0, 0, 0, 0, 0, 0),
        ('Andelfingen', 30, 0, 0, 0, 0, 0, 0)
    ]
    assert report.total() == (0, 0, 0, 0, 0, 0)


def test_report_forms_by_municipality(session):
    def _report(start, end, municipality):
        query = session.query(Municipality).filter_by(name=municipality)
        return ReportFormsByMunicipality(
            session, start=start, end=end, municipality_id=query.one().id
        )

    add_report_data(session)

    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Adlikon')
    assert report.query().all() == [('Adlikon', 21, 1, 2, 3, 6)]
    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Aesch')
    assert report.query().all() == [('Aesch', 241, 1, 2, 3, 6)]

    report = _report(date(2019, 1, 2), date(2019, 1, 3), 'Adlikon')
    assert report.query().all() == [('Adlikon', 21, 3, 2, 1, 6)]
    report = _report(date(2019, 1, 2), date(2019, 1, 3), 'Aesch')
    assert report.query().all() == [('Aesch', 241, 0, 10, 0, 10)]
    report = _report(date(2019, 1, 2), date(2019, 1, 3), 'Altikon')
    assert report.query().all() == [('Altikon', 211, 1, 2, 3, 6)]
    report = _report(date(2019, 1, 2), date(2019, 1, 3), 'Andelfingen')
    assert report.query().all() == [('Andelfingen', 30, 1, 2, 3, 6)]

    report = _report(date(2019, 1, 4), date(2019, 1, 5), 'Aesch')
    assert report.query().all() == [('Aesch', 241, 0, 0, 0, 0)]
    report = _report(date(2019, 1, 4), date(2019, 1, 5), 'Andelfingen')
    assert report.query().all() == [('Andelfingen', 30, 0, 0, 0, 0)]


def test_report_delivery(session):
    def _report(start, end, municipality):
        query = session.query(Municipality).filter_by(name=municipality)
        return ReportBoxesAndFormsByDelivery(
            session, start=start, end=end, type=None,
            municipality_id=query.one().id
        )

    add_report_data(session)

    # 2019-01-01
    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Adlikon')
    assert report.query().all() == [(date(2019, 1, 1), 1, 1, 2, 3, 6, 4, 4)]
    assert report.total() == (0, 0, 1, 2, 3, 6, 4, 4)

    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Aesch')
    assert report.query().all() == [(date(2019, 1, 1), 3, 1, 2, 3, 6, 4, 4)]
    assert report.total() == (0, 0, 1, 2, 3, 6, 4, 4)

    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Altikon')
    assert report.query().all() == []
    assert report.total() == (0, 0, 0, 0, 0, 0, 0, 0)

    report = _report(date(2019, 1, 1), date(2019, 1, 1), 'Andelfingen')
    assert report.query().all() == []
    assert report.total() == (0, 0, 0, 0, 0, 0, 0, 0)

    # 2019-01-01 ... 2019-01-05
    report = _report(date(2019, 1, 1), date(2019, 1, 5), 'Adlikon')
    assert report.query().all() == [
        (date(2019, 1, 1), 1, 1, 2, 3, 6, 4, 4),
        (date(2019, 1, 2), 2, 3, 2, 1, 6, 1, 1)
    ]
    assert report.total() == (0, 0, 4, 4, 4, 12, 5, 5)

    report = _report(date(2019, 1, 1), date(2019, 1, 5), 'Aesch')
    assert report.query().all() == [
        (date(2019, 1, 1), 3, 1, 2, 3, 6, 4, 4),
        (date(2019, 1, 3), 4, 0, 10, 0, 10, 0, None)
    ]
    assert report.total() == (0, 0, 1, 12, 3, 16, 4, 4)

    report = _report(date(2019, 1, 1), date(2019, 1, 5), 'Altikon')
    assert report.query().all() == [
        (date(2019, 1, 2), 5, 1, 2, 3, 6, 4, 4)
    ]
    assert report.total() == (0, 0, 1, 2, 3, 6, 4, 4)

    report = _report(date(2019, 1, 1), date(2019, 1, 5), 'Andelfingen')
    assert report.query().all() == [
        (date(2019, 1, 2), 6, 1, 2, 3, 6, 4, 4)
    ]
    assert report.total() == (0, 0, 1, 2, 3, 6, 4, 4)


def test_notification(session):

    class Identity():
        application_id = 'wtfs'
        userid = 'admin'

    class Request():
        def __init__(self, session):
            self.identity = Identity()
            self.session = session

    notification = Notification.create(
        Request(session),
        title="Lorem ipsum",
        text="Lorem ipsum dolor sit amet."
    )

    notification = session.query(Notification).one()
    assert notification.title == "Lorem ipsum"
    assert notification.text == "Lorem ipsum dolor sit amet."
    assert notification.channel_id == "wtfs"
    assert notification.owner == "admin"
    assert notification.type == "wtfs_notification"


def test_invoice(session):
    session.add(PaymentType(name='normal', _price_per_quantity=700))
    session.add(PaymentType(name='spezial', _price_per_quantity=850))

    session.add(
        Municipality(
            name='Adlikon',
            address_supplement='Finkenweg',
            gpn_number=8882255,
            bfs_number=21,
            payment_type='normal'
        )
    )
    session.add(
        Municipality(
            name='Aesch',
            address_supplement='Meisenweg',
            gpn_number=9993344,
            bfs_number=241,
            payment_type='spezial'
        )
    )
    session.flush()
    adlikon = session.query(Municipality).filter_by(name='Adlikon').one()
    aesch = session.query(Municipality).filter_by(name='Aesch').one()

    session.add(
        ScanJob(
            type='normal',
            municipality_id=adlikon.id,
            dispatch_date=date(2018, 12, 10),
            dispatch_boxes=1,
            dispatch_tax_forms_current_year=10,
            dispatch_tax_forms_last_year=20,
            dispatch_tax_forms_older=30,
            dispatch_single_documents=40,
            dispatch_cantonal_tax_office=5,
            dispatch_cantonal_scan_center=4,
            return_date=date(2018, 12, 20),
            return_boxes=1,
            return_tax_forms_current_year=9,
            return_tax_forms_last_year=18,
            return_tax_forms_older=27,
            return_single_documents=36,
            return_unscanned_tax_forms_current_year=1,
            return_unscanned_tax_forms_last_year=2,
            return_unscanned_tax_forms_older=3,
            return_unscanned_single_documents=4,
        )
    )
    session.add(
        ScanJob(
            type='normal',
            municipality_id=adlikon.id,
            dispatch_date=date(2019, 1, 10),
            dispatch_boxes=2,
            dispatch_tax_forms_current_year=5,
            dispatch_tax_forms_last_year=5,
            dispatch_tax_forms_older=0,
            dispatch_single_documents=2,
            dispatch_cantonal_tax_office=0,
            dispatch_cantonal_scan_center=0,
            return_date=date(2019, 2, 2),
            return_boxes=2,
            return_tax_forms_current_year=5,
            return_tax_forms_last_year=5,
            return_tax_forms_older=0,
            return_single_documents=2,
            return_unscanned_tax_forms_current_year=0,
            return_unscanned_tax_forms_last_year=0,
            return_unscanned_tax_forms_older=0,
            return_unscanned_single_documents=0,
        )
    )
    session.add(
        ScanJob(
            type='normal',
            municipality_id=aesch.id,
            dispatch_date=date(2019, 1, 7),
            dispatch_boxes=1,
            dispatch_tax_forms_current_year=10,
            dispatch_tax_forms_last_year=10,
            dispatch_tax_forms_older=10,
            dispatch_single_documents=10,
            dispatch_cantonal_tax_office=1,
            dispatch_cantonal_scan_center=1,
            return_date=date(2019, 1, 10),
            return_boxes=1,
            return_tax_forms_current_year=7,
            return_tax_forms_last_year=7,
            return_tax_forms_older=7,
            return_single_documents=7,
            return_unscanned_tax_forms_current_year=3,
            return_unscanned_tax_forms_last_year=3,
            return_unscanned_tax_forms_older=3,
            return_unscanned_single_documents=3,
        )
    )
    session.flush()

    with freeze_time("2019-12-31 08:07:06"):
        header = (
            'Aktuelles Datum (YYYYMMDD1),'
            'Aktuelles Datum (DD.MM.YYYY),'
            'Download Zeitpunkt (HH.MIN.SEK),'
            'GPN-Nr.,'
            'Positionsnummer (Hochgezählt),'
            'GPN-Nr.,'
            'Betreff,'
            'Positionsnummer (Hochgezählt),'
            'Erstellungsdatum,Erstellungsdatum,Erstellungsdatum,'
            'GPN-Nr.,'
            'Total StE,'
            'Lieferscheinnummer,'
            'Preis pro Menge exkl. MWST,Preis pro Menge exkl. MWST,'
            'Fixe Preiseiheit (immer „1“),'
            'Erstellungsdatum,'
            'Adressszusatz der Gemeinde,'
            'CS2 Benutzer,'
            'Kostenstelle,'
            'Erlöskonto'
            '\r\n'
        )

        invoice = Invoice(session)

        def export():
            file = StringIO()
            invoice.export(file)
            file.seek(0)
            return file.read()

        # All
        invoice.municipality_id = None
        invoice.from_date = None
        invoice.to_date = None
        invoice.cs2_user = '123456'
        invoice.subject = 'Rechnungen'
        invoice.accounting_unit = '99999'
        invoice.revenue_account = '987654321'
        invoice.vat = None
        assert export() == header + (
            '201912311,31.12.2019,08.07.06,'
            '8882255,'
            '1,'
            '8882255,'
            'Rechnungen,'
            '1,'
            '31.12.2019,31.12.2019,31.12.2019,'
            '8882255,'
            '54000,'
            '1,'
            '70000000,70000000,'
            '1,'
            '31.12.2019,'
            'Finkenweg,'
            '123456,'
            '99999,'
            '987654321'
            '\r\n'
        ) + (
            '201912311,31.12.2019,08.07.06,'
            '9993344,'
            '2,'
            '9993344,'
            'Rechnungen,'
            '2,'
            '31.12.2019,31.12.2019,31.12.2019,'
            '9993344,'
            '21000,'
            '3,'
            '85000000,85000000,'
            '1,'
            '31.12.2019,'
            'Meisenweg,'
            '123456,'
            '99999,'
            '987654321'
            '\r\n'
        ) + (
            '201912311,31.12.2019,08.07.06,'
            '8882255,'
            '3,'
            '8882255,'
            'Rechnungen,'
            '3,'
            '31.12.2019,31.12.2019,31.12.2019,'
            '8882255,'
            '10000,'
            '2,'
            '70000000,70000000,'
            '1,'
            '31.12.2019,'
            'Finkenweg,'
            '123456,'
            '99999,'
            '987654321'
            '\r\n'
        )

        # Filter
        invoice.municipality_id = adlikon.id
        invoice.from_date = date(2019, 1, 1)
        invoice.to_date = date(2019, 1, 7)
        invoice.cs2_user = '123456'
        invoice.subject = 'Rechnungen 1.1-7.1'
        invoice.accounting_unit = '99999'
        invoice.revenue_account = '987654321'
        invoice.vat = None
        assert export() == header

        invoice.from_date = date(2019, 1, 7)
        invoice.to_date = date(2019, 1, 14)
        assert export() == header + (
            '201912311,31.12.2019,08.07.06,'
            '8882255,'
            '1,'
            '8882255,'
            'Rechnungen 1.1-7.1,'
            '1,'
            '31.12.2019,31.12.2019,31.12.2019,'
            '8882255,'
            '10000,'
            '2,'
            '70000000,70000000,'
            '1,'
            '31.12.2019,'
            'Finkenweg,'
            '123456,'
            '99999,'
            '987654321'
            '\r\n'
        )



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.wtfs import _
from onegov.wtfs.layouts.default import DefaultLayout


class ReportLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Report")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.request.link(self.model))
        ]

    @cached_property
    def dates(self):
        return "{}-{}".format(
            self.format_date(self.model.start, 'date'),
            self.format_date(self.model.end, 'date')
        )


class SpecificReportBaseLayout(DefaultLayout):

    @cached_property
    def subtitle(self):
        return "{}-{}".format(
            self.format_date(self.model.start, 'date'),
            self.format_date(self.model.end, 'date')
        )

    @cached_property
    def editbar_links(self):
        result = []
        result.append(
            Link(
                text=_("Print"),
                url='#',
                attrs={
                    'class': 'print-icon',
                    'onclick': 'window.print();return false;'
                }
            )
        )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Report"), self.report_url),
            Link(self.title, "#"),
            Link(self.subtitle, self.request.link(self.model)),
        ]


class ReportBoxesLayout(SpecificReportBaseLayout):

    @cached_property
    def title(self):
        return _("Report boxes")


class ReportBoxesAndFormsLayout(SpecificReportBaseLayout):

    @cached_property
    def title(self):
        return _("Report boxes and forms")


class ReportFormsByMunicipalityLayout(SpecificReportBaseLayout):

    @cached_property
    def title(self):
        return _("Report forms")

    @cached_property
    def subtitle(self):
        return "{} {}-{}".format(
            self.model.municipality_name,
            self.format_date(self.model.start, 'date'),
            self.format_date(self.model.end, 'date')
        )


class ReportBoxesAndFormsByDeliveryLayout(SpecificReportBaseLayout):

    @cached_property
    def title(self):
        return _("Report boxes and forms by delivery")

    @cached_property
    def subtitle(self):
        return "{} ({}) {}-{}".format(
            self.model.municipality.name,
            self.model.municipality.bfs_number,
            self.format_date(self.model.start, 'date'),
            self.format_date(self.model.end, 'date')
        )



from onegov.wtfs.forms.daily_list import DailyListSelectionForm
from onegov.wtfs.forms.invoice import CreateInvoicesForm
from onegov.wtfs.forms.municipality import DeleteMunicipalityDatesForm
from onegov.wtfs.forms.municipality import ImportMunicipalityDataForm
from onegov.wtfs.forms.municipality import MunicipalityForm
from onegov.wtfs.forms.municipality import MunicipalityIdSelectionForm
from onegov.wtfs.forms.notification import NotificationForm
from onegov.wtfs.forms.payment_type import PaymentTypesForm
from onegov.wtfs.forms.report import ReportSelectionForm
from onegov.wtfs.forms.scan_job import AddScanJobForm
from onegov.wtfs.forms.scan_job import EditScanJobForm
from onegov.wtfs.forms.scan_job import ScanJobsForm
from onegov.wtfs.forms.scan_job import UnrestrictedScanJobForm
from onegov.wtfs.forms.scan_job import UnrestrictedScanJobsForm
from onegov.wtfs.forms.user import UnrestrictedUserForm
from onegov.wtfs.forms.user import UserForm


__all__ = (
    'AddScanJobForm',
    'CreateInvoicesForm',
    'DailyListSelectionForm',
    'DeleteMunicipalityDatesForm',
    'EditScanJobForm',
    'ImportMunicipalityDataForm',
    'MunicipalityForm',
    'MunicipalityIdSelectionForm',
    'NotificationForm',
    'PaymentTypesForm',
    'ReportSelectionForm',
    'ScanJobsForm',
    'UnrestrictedScanJobForm',
    'UnrestrictedScanJobsForm',
    'UnrestrictedUserForm',
    'UserForm',
)
